""" CLI for `cedar auth`
"""
import logging
import os
from pathlib import Path

import click

from . import options

logger = logging.getLogger(__name__)


# Help links
DOCS_AUTH_GCS = ''
DOCS_AUTH_GDRIVE = ''


@click.group('auth', help='Log in to GEE pre-ARD services')
@click.pass_context
def group_auth(ctx):
    pass


@group_auth.command('ee', help='Test logging into to GEE service')
@click.pass_context
def auth_gee(ctx):
    """ Test logging into the Google Earth Engine

    If it doesn't work, please make sure you've authenticated by running
    >>> earthengine authenticate
    """
    from cedar.utils import load_ee

    try:
        ee = load_ee(True)
    except Exception as e:
        click.echo('Could not load and initialize the Earth Engine API')
        click.echo(e)
        raise click.Abort()
    else:
        click.echo(options.STYLE_INFO('Authenticated'))


@group_auth.command('gdrive', help='Login to use Google Drive')
@click.option('--client_secrets_file', help='OAuth2 "client secrets" file',
              type=click.Path(dir_okay=False, resolve_path=True, exists=True))
@click.option('--credentials_file', help='OAuth2 credentials',
              type=click.Path(dir_okay=False, resolve_path=True))
@options.opt_browser
@click.pass_context
def auth_gdrive(ctx, client_secrets_file, credentials_file, browser):
    """ Login to the Google Drive API service using OAuth2 credentials

    Useful for:
        1. Initially authenticating using a ``client_secrets_file`` so that
           user authentication credentials may be written and used later
           (in your configuration file)
        2. Checking to make sure things are working
    """
    from cedar.stores import gdrive
    from cedar.config import Config

    # CLI overrides info from config, if passed
    config = options.fetch_config(ctx, False)
    if config:
        cfg = config.get('gdrive', {})
        client_secrets_file = _override_from_cfg(
            client_secrets_file, 'client_secrets_file', cfg)
        credentials_file = _override_from_cfg(
            credentials_file, 'credentials_file', cfg)

    # Create/refresh credentials
    creds, creds_file = gdrive.get_credentials(
        client_secrets_file=client_secrets_file,
        credentials_file=credentials_file,
        no_browser=not browser)

    # Check we can build a service
    service = gdrive.build_gdrive_service(creds)
    click.echo(f'Authenticated using credentials file {creds_file}')


@group_auth.command('gcs', help='Login to use Google Cloud Storage')
@click.option('--service_account_file', type=click.Path(resolve_path=True),
              envvar='GOOGLE_APPLICATION_CREDENTIALS',
              help='Google service account file for GCS')
@click.option('--project', envvar='GOOGLE_CLOUD_PROJECT',
              type=str, help='GCS project to use')
@click.pass_context
def auth_gcs(ctx, service_account_file, project):
    """ Test authenticating to Google Cloud Storage
    """
    # TODO: try to build gcs service
    from cedar.stores import gcs

    config = options.fetch_config(ctx, False)
    config_gcs = config.get('gcs', {})

    # Service account file preference: CLI option > config
    service_account_file_ = (service_account_file or
                             config_gcs.get('service_account_file', None))
    envvar_creds = os.environ.get('GOOGLE_APPLICATION_CREDENTIALS', None)
    if not service_account_file_ and not envvar_creds:
        raise click.BadParameter(
            'Must specify `service_account_file` in `gcs` section of '
            'config file', param_hint='service_account_file')

    project_ = project or config_gcs.get('project', None)
    if not project_:
        raise click.BadParameter(
            'Must specify `project` in `gcs` section of config file',
            param_hint='project')

    try:
        client = gcs.build_gcs_client(service_account_file_, project=project_)
    except Exception as e:
        click.echo('Could not authenticate to Google Cloud Storage')
        raise e
    else:
        click.echo('Authenticated using service account file '
                   f'"{service_account_file}" and project "{project}"')
        if service_account_file or project:
            click.echo('Make sure to add this information to the `gcs` '
                       'section of your configuration file')


@group_auth.command('clear', help='Clear credentials')
@click.option('--yes', is_flag=True, help='Assume YES to deleting files')
@click.pass_context
def clear(ctx, yes):
    """ Delete credentials files
    """
    config = options.fetch_config(ctx, False)

    to_clear = []

    config_gdrive = config.get('gdrive', {})
    to_clear.extend(_get_gdrive_creds(config_gdrive))

    # TODO: get GCS cred files
    config_gcs = config.get('gcs', {})

    removed = []
    for cred_file in to_clear:
        if cred_file is not None:
            cred_file = Path(cred_file)
            msg = f'Do you want to delete credential file "{str(cred_file)}"?'
            if cred_file.exists() and (yes or click.confirm(msg)):
                try:
                    cred_file.unlink()
                except FileNotFoundError:  # race condition
                    pass
                else:
                    removed.append(cred_file)
    if removed:
        for cred_file in removed:
            click.echo(f'Removed {cred_file}')
    else:
        click.echo('Did not delete any credential files')


def _get_gdrive_creds(config):
    # Remove Google Drive credentials
    try:
        from cedar.stores import gdrive
    except ImportError:
        return (None, None)

    secrets = config.get('client_secrets_file', None)
    creds = config.get('credentials_file', None)

    return gdrive.find_credentials(secrets, creds)


def _override_from_cfg(value, key, cfg):
    if value is None and key in cfg:
        logger.debug(f'Using `{key}` from `config_file`')
        return cfg[key]
    else:
        return value



""" Common utility functions
"""
import logging
import os
from pathlib import Path
import stat

logger = logging.getLogger(__name__)


def get_file(*filenames, exists=True):
    """ Return first good filename

    Parameters
    ----------
    filenames : Sequence[str or Path]
        Filenames to check
    exists : bool, optional
        Require that the filename correspond to an existing
        file on disk

    Returns
    -------
    Path or None
        A Path of the first good filename, or None if they're all bad
    """
    for filename in filenames:
        if filename:  # not null
            filename = Path(filename)
            if exists:  # if we need to check...
                if filename.exists():
                    logger.debug(f'Found filename "{filename}"')
                    return Path(filename)
                elif Path.cwd().joinpath(filename).exists():
                    logger.debug(f'Found filename "{filename}" in CWD')
                    return Path.cwd().joinpath(filename)
                else:
                    logger.debug(f'File not found at "{filename}"')
            else:  # otherwise just return first non-null
                return Path(filename)
    return None


def set_file_urw(filename):
    """ Sets a file to user read/write only

    Parameters
    ----------
    filename : str
        File to modify
    """
    os.chmod(str(filename), stat.S_IREAD | stat.S_IWRITE)


def affine_to_str(transform):
    """ Return a string representatin of an affine.Affine transform
    """
    return ','.join(map(str, transform[:6]))


def load_ee(initialize=True):
    """ Import and initialize the EE API, handling errors

    Parameters
    ----------
    initialize : bool, optional
        Try to initialize the EE API, or not

    Returns
    -------
    object
        Earth Engine API module ``ee``

    Raises
    ------
    ImportError
        Raised if the Earth Engine cannot be imported
    ee.ee_exceptions.EEException
        Raised if the Earth Engine cannot be initialized, typically because
        you haven't authenticated yet.
    Exception
        Passes other exceptions caused by :py:func:`ee.Initialize`
    """
    try:
        import ee as ee_api
    except ImportError as ie:
        docs = 'https://developers.google.com/earth-engine/python_install'
        pip_info = '`pip` ("pip install earthengine-api")'
        conda_info = '`conda` ("conda install -c conda-forge earthengine-api")'
        raise ImportError(
            f'Cannot import the Earth Engine API. Please install the package, '
            f'which is available through {pip_info} or {conda_info}. Visit '
            f'"{docs}" for more information.'
        )
    else:
        if initialize:
            try:
                ee_api.Initialize()
            except ee_api.ee_exception.EEException as eee:
                raise eee
            except Exception:
                logger.exception(
                    'Could not initialize EarthEngine API. Depending on the '
                    'error, this might be caused by network access issues.'
                )
                raise
            else:
                return ee_api
        else:
            return ee_api


def get_ee_tasks():
    """ Return GEE tasks (task ID: task)

    Returns
    -------
    dict[str, ee.batch.task.Task]
        GEE tasks
    """
    ee = load_ee(False)
    return {task.id: task for task in ee.batch.Task.list()}


# =============================================================================
# Earth Engine filters
def serialize_filter(ee_filter):
    """ Serialize an Earth Engine filter
    """
    ee = load_ee(False)
    return ee.serializer.encode(ee_filter)


def create_filters(filters):
    """ Convert a list of filters/filter descriptions to List[ee.Filter]

    Parameters
    ----------
    filters : List[dict]
        Earth Engine filters serialized to dict

    Returns
    -------
    List[ee.Filter]
        Filters as ee.Filter
    """
    ee = load_ee(False)
    filters = []
    for filter_ in filters:
        if isinstance(filter_, ee.Filter):
            filters.append(filter_)
        else:
            filters.append(dict_to_filter(**filter_))
    return filters


def dict_to_filter(function, **kwds):
    """ Convert serialized form of filter to a ee.Filter

    Parameters
    ----------
    function : str
        Name of filter. Should be a (static) method of ``ee.Filter``
    kwds
        Keyword arguments to pass to for the filter construction. These
        will depend on the filter in question

    Returns
    -------
    ee.Filter
        Earth Engine filter
    """
    ee = load_ee(False)
    static_meth = getattr(ee.Filter, function)
    return static_meth(**kwds)



""" CLI for opening a Python interactive terminal
"""
import click


@click.command('console', short_help='Enter cedar interactive console')
@click.option('--ipython', is_flag=True, help='Use IPython console')
@click.pass_context
def console(ctx, ipython):
    """ Enter interactive Python shell with a Config
    """
    from . import options
    config = options.fetch_config(ctx)

    banner = 'cedar interactive console'
    local = {'config': config}

    if ipython:
        import IPython
        IPython.InteractiveShell.banner1 = banner
        IPython.start_ipython(argv=[], user_ns=local)
    else:
        import code
        code.interact(banner, local=local)



""" CLI for checking GEE tasks
"""
from collections import defaultdict
import logging

import click

from . import options


@click.group('gee', short_help='Check status GEE tasks')
@click.pass_context
def group_gee(ctx):
    pass


@group_gee.command('tasks', short_help='List Google Earth Engine tasks')
@click.option('--list', 'list_tasks', is_flag=True,
              help='List information for all tasks')
@click.pass_context
def tasks(ctx, list_tasks):
    """ Get info about Google Earth Engine tasks
    """
    from cedar.utils import get_ee_tasks, load_ee

    ee = load_ee(True)
    tasks = get_ee_tasks()

    # Summarize
    totals = defaultdict(lambda: 0)
    for task in tasks.values():
        totals[task.state] += 1
    click.echo('Task summary:')
    for state, count in totals.items():
        click.echo(f'    {state}: {count}')

    # Print out verbose info
    if list_tasks:
        click.echo('Tasks:')
        for id_, task in tasks.items():
            click.echo(f'{id_} - {task.state} - {task.task_type}=>'
                       f'{task.config["description"]}')



""" CLI tools for submitting "pre-ARD" orders to GEE
"""
import itertools
import logging

import click

from stems.cli import options as cli_options

from .. import defaults
from . import options

logger = logging.getLogger(__name__)


def _check_imgcol(ctx, image_collection):
    from cedar.sensors import CREATE_ARD_COLLECTION
    for collection in image_collection:
        if collection not in CREATE_ARD_COLLECTION:
            _, imgcol_param = options.fetch_param(ctx, 'image_collection')
            img_cols = ", ".join([f'"{k}"' for k in CREATE_ARD_COLLECTION])
            raise click.BadParameter(
                f'Unknown image collection "{collection}". '
                f'Must be one of {img_cols}',
                ctx=ctx, param=imgcol_param)
    return image_collection


@click.command('submit', short_help='Submit "pre-ARD" data processing tasks')
@click.argument('image_collection', nargs=-1, type=str, required=True,
                callback=_check_imgcol)
@click.option('--index', '-i', nargs=2, type=(int, int), multiple=True,
  help='TileGrid (row, col) index(es) to submit')
@click.option('--row', '-r', type=int, multiple=True,
  help='TileGrid row(s) to submit. Use in conjunction with `--col`')
@click.option('--col', '-c', type=int, multiple=True,
  help='TileGrid col(s) to submit. Use in conjunction with `--row`')
@click.option('--period_start', callback=cli_options.cb_time,
  help='Starting time period for submission')
@click.option('--period_end', callback=cli_options.cb_time,
  help='Ending time period for submission')
@click.option('--period_freq', type=str, default=defaults.PREARD_FREQ,
  help='Split start/end time into periods of this frequency')
@cli_options.opt_date_format
@click.pass_context
def submit(ctx, image_collection, index, row, col,
           period_start, period_end, period_freq, date_format):
    """ Submit "pre-ARD" processing orders and create tracking metadata
    """
    from cedar.exceptions import EmptyOrderError
    from cedar.sensors import CREATE_ARD_COLLECTION
    from cedar.utils import load_ee

    # Need index OR (row AND col) -- blame index param if it goes wrong
    _, index_param = options.fetch_param(ctx, 'index')
    if index:
        if row or col:
            raise click.BadParameter(
                'Cannot use `--index` with `--row` or `--col`',
                ctx=ctx, param=index_param)
    elif row and col:
        if index:
            raise click.BadParameter(
                'Cannot use `--row` and `--col` with `--index`',
                ctx=ctx, param=index_param)
        index = list(itertools.product(row, col))
    else:
        raise click.BadParameter(
            'Must specify `--index` OR both `--row` and `--col`',
            ctx=ctx, param=index_param)

    # Eventually we'll support not having a period_start/period_end, but not yet
    if not period_start or not period_end:
        name = f'period_{"start" if not period_start else "end"}'
        param = options.fetch_param(ctx, name)[1]
        raise click.BadParameter(f'Must pass {name}ing period',
                                 ctx=ctx, param=param)

    # Check that we know about the image collection
    for collection in image_collection:
        if collection not in CREATE_ARD_COLLECTION:
            raise KeyError(f'Unknown image collection "{collection}"')

    # Get parse(d) config and build tracker
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    # Login to EE
    ee = load_ee(True)

    msg = [
        f'Tiles: {", ".join([f"h{c:04d}v{r:04d}" for r, c in index])}',
        f'Collections: {", ".join(image_collection)}',
        f'Time period: {period_start.isoformat()} - {period_end.isoformat()}',
        f'At frequency: {period_freq}'
    ]
    click.echo('Submitting preARD tasks for:')
    click.echo("\n".join([f"    {s}" for s in msg]))

    # Submit!
    try:
        tracking_info_name, tracking_info_id = tracker.submit(
            image_collection,
            index,
            period_start, period_end,
            period_freq=period_freq
        )
        click.echo('Wrote job tracking to store object named '
                   f'"{tracking_info_name}" ({tracking_info_id})')
    except EmptyOrderError as ece:
        click.echo(repr(ece))
        click.echo('Did not find data to order. Check your search parameters')
        raise click.Abort()
    except Exception as e:
        click.echo('Unknown error occurred. See exception printed below')
        raise e



""" Functions for dealing with Landsat data on GEE
"""
import datetime as dt
import logging

import ee

from ..exceptions import EmptyCollectionError
from . import common

logger = logging.getLogger(__name__)

# Renaming stuff
BANDS_COMMON = ['blue', 'green', 'red', 'nir',
                'swir1', 'swir2', 'thermal', 'pixel_qa']

BANDS_LT4 = ['B1', 'B2', 'B3', 'B4', 'B5', 'B7',  'B6', 'pixel_qa']
BANDS_LT5 = BANDS_LT4.copy()
BANDS_LE7 = ['B1', 'B2', 'B3', 'B4', 'B5', 'B7',  'B6', 'pixel_qa']
BANDS_LC8 = ['B2', 'B3', 'B4', 'B5', 'B6', 'B7', 'B10', 'pixel_qa']

BANDS = {
    'COMMON': BANDS_COMMON,
    'LANDSAT/LT04/C01/T1_SR': BANDS_LT4,
    'LANDSAT/LT05/C01/T1_SR': BANDS_LT5,
    'LANDSAT/LE07/C01/T1_SR': BANDS_LE7,
    'LANDSAT/LC08/C01/T1_SR': BANDS_LC8,
}

#: dict[str, Number]: NoDataValues for Landsat collections
NODATA = {
    'LANDSAT/LT04/C01/T1_SR': -9999,
    'LANDSAT/LT05/C01/T1_SR': -9999,
    'LANDSAT/LE07/C01/T1_SR': -9999,
    'LANDSAT/LC08/C01/T1_SR': -9999,
}


_T1_SR_METADATA = [
    'CLOUD_COVER',
    'CLOUD_COVER_LAND',
    'EARTH_SUN_DISTANCE',
    'ESPA_VERSION',
    'GEOMETRIC_RMSE_MODEL',
    'GEOMETRIC_RMSE_MODEL_X',
    'GEOMETRIC_RMSE_MODEL_Y',
    'LANDSAT_ID',
    'LEVEL1_PRODUCTION_DATE',
    'SATELLITE',
    'SENSING_TIME',
    'SOLAR_AZIMUTH_ANGLE',
    'SOLAR_ZENITH_ANGLE',
    'SR_APP_VERSION',
    'WRS_PATH',
    'WRS_ROW',
    'system:id',
    'system:time_start',
    'system:version'
]
_T1_SR_METADATA_LT04 = ['IMAGE_QUALITY']
_T1_SR_METADATA_LT05 = ['IMAGE_QUALITY']
_T1_SR_METADATA_LE07 = ['IMAGE_QUALITY']
_T1_SR_METADATA_LC08 = ['IMAGE_QUALITY_OLI']
METADATA = {
    'LANDSAT/LT04/C01/T1_SR': _T1_SR_METADATA + _T1_SR_METADATA_LT04,
    'LANDSAT/LT05/C01/T1_SR': _T1_SR_METADATA + _T1_SR_METADATA_LT05,
    'LANDSAT/LE07/C01/T1_SR': _T1_SR_METADATA + _T1_SR_METADATA_LE07,
    'LANDSAT/LC08/C01/T1_SR': _T1_SR_METADATA + _T1_SR_METADATA_LC08,
}


def create_ard(collection, tile, date_start, date_end, filters=None,
               validate=False):
    """ Create an ARD :py:class:`ee.Image`

    Parameters
    ----------
    collection : str
        GEE image collection name
    tile : stems.gis.grids.Tile
        STEMS TileGrid tile
    date_start : dt.datetime
        Starting period
    date_end : dt.datetime
        Ending period
    filters : Sequence[ee.Filter], optional
        Additional filters to apply over image collection
    validate : bool, optional
        Perform validity checks at cost of submission speed (runs ``.getInfo``
        on metadata, requiring us to wait on client-server communication)

    Returns
    -------
    ee.Image
        "ARD" image from collection with all observations within period
    Sequence[dict]
        Metadata, one dict per image
    """
    # TODO: convert system:time_start to datetime/strftime
    assert isinstance(date_start, dt.datetime)
    assert isinstance(date_end, dt.datetime)

    # Get collection
    if isinstance(collection, ee.ImageCollection):
        imgcol = collection
        collection = imgcol.get('system:id').getInfo()
    else:
        imgcol = ee.ImageCollection(collection)

    if not collection in BANDS.keys():
        raise KeyError(f'Image collection "{collection}" is unsupported')

    # Find images in tile
    imgcol = common.filter_collection_tile(imgcol, tile)

    # For each unique date of imagery in this image collection covering the tile
    imgcol = common.filter_collection_time(imgcol, date_start, date_end)

    # Apply additional filters
    if filters:
        logger.debug(f'Applying {len(filters)} filters over collection')
        imgcol = imgcol.filter(filters)

    # Select and rename bands
    imgcol = imgcol.select(BANDS[collection], BANDS['COMMON'])

    # Find number of unique observations (or, uniquely dated)
    imgcol_udates = common.get_collection_uniq_dates(imgcol)
    n_images = len(imgcol_udates)
    if n_images == 0:
        raise EmptyCollectionError(f'Found 0 images for "{collection}" between '
                                   f'{date_start}-{date_end}')

    # Loop over unique dates, making mosaics to eliminate north/south if needed
    logger.debug(f'Creating ARD for {n_images} images')
    prepped = []
    for udate in sorted(imgcol_udates):
        # Prepare and get metadata for unique date
        img, meta = _prep_collection_image(imgcol, collection, udate,
                                           validate=validate)
        # Add image and metadata
        prepped.append((img, meta))

    # Unpack
    images, _ = list(zip(*prepped))

    # Re-create as collection and turn to bands (n_image x bands_per_image)
    tile_col = ee.ImageCollection.fromImages(images)
    tile_bands = tile_col.toBands().toInt16()

    # Remove mask
    nodata = NODATA[collection]
    tile_bands_unmasked = tile_bands.unmask(nodata)

    # Get all image metadata at once (saves time back and forth)
    images_metadata = list(ee.List(_).getInfo())


    # Create overall metadata
    band_names = list(imgcol.first().bandNames().getInfo())
    metadata = {
        'bands': band_names,
        'nodata': nodata,
        'images': images_metadata
    }

    return tile_bands_unmasked, metadata


def _imgcol_metadata(imgcol, keys):
    """ Return metadata for Landsat image collection
    """
    def inner(img, previous):
        meta = common.object_metadata(img, keys)
        previous_ = ee.List(previous)
        return ee.List(previous_.add(meta))

    meta = imgcol.iterate(inner, ee.List([]))
    return meta


def _prep_collection_image(imgcol, collection, date, validate=False):
    """ Prepare an image for and ``date`` from an ImageCollection
    """
    # Filter for this date (day <-> day+1)
    date_end = (date + dt.timedelta(days=1))
    imgcol_ = common.filter_collection_time(imgcol, date, date_end)

    if validate:
        # Check to make sure just 1 unique date
        _ = common.get_collection_uniq_dates(imgcol_)
        assert len(_) == 1

    # Prepare all images in this collection (i.e., 1 or 2, depending on overlap)
    img = imgcol_.mosaic()

    # Get metadata from each image in new, potentially mosaiced ``img``
    keys = METADATA[collection]
    meta = _imgcol_metadata(imgcol_, keys)

    return img, meta



""" Common functions for dealing with GEE
"""
import datetime as dt

import ee
import shapely.geometry


# ==============================================================================
# Metadata handling
def object_metadata(obj, keys):
    """ Return metadata from some EE object as dict (not evaluated)

    Parameters
    ----------
    image : object
        EarthEngine object (from ``ee``), typically an ``ee.Image``
    keys : Sequence[str]
        Metadata to retrieve

    Returns
    -------
    ee.Dictionary
        Dictionary of metadata
    """
    d = ee.Dictionary()
    for key in keys:
        d = d.set(key, obj.get(key))
    return d


def collection_metadata(collection, key):
    """ Return a ee.List of metadata from each item in ``collection``

    Parameters
    ----------
    collection : ee.ImageCollection
        GEE collection
    key : str
        Metadata key

    Returns
    -------
    ee.List
        List of metadata per item in ``collection``
    """
    def inner(img, previous):
        v = ee.Image(img).get(key)
        previous_ = ee.List(previous)
        return ee.List(previous_.add(v))
    return collection.iterate(inner, ee.List([]))


# ==============================================================================
# Time
def get_collection_dates(imgcol):
    def inner(image, previous):
        date = ee.Image(image).date().format('YYYY-MM-dd')
        previous_ = ee.List(previous)
        return ee.List(previous_.add(date))
    dates = imgcol.iterate(inner, ee.List([])).getInfo()
    return [dt.datetime.strptime(d, '%Y-%m-%d') for d in dates]


def get_collection_uniq_dates(col):
    return list(sorted(set(get_collection_dates(col))))


def filter_collection_time(col, d_start, d_end):
    # ee.Date should parse from str or datetime
    d_start_ = ee.Date(d_start)
    d_end_ = ee.Date(d_end)
    col_ = col.filterDate(d_start_, d_end_)
    return col_


# =============================================================================
# Tile related functions
def filter_collection_tile(col, tile):
    geom_ee = tile_geom(tile)
    return col.filterBounds(geom_ee)


def tile_geom(tile):
    # Buffer slightly inward so we don't get "touching" but not overlapping
    geom = shapely.geometry.mapping(tile.bbox.buffer(-1e-3))
    geom_ee = ee.Geometry(geom, opt_proj=tile.crs.wkt)
    return geom_ee


""" CLI for converting "pre-ARD" to ARD data cubes
"""
import logging
import os.path
from pathlib import Path

import click

from stems.cli import options as cli_options

from . import options


@click.command('convert',
               short_help='Convert downloaded "pre-ARD" data to ARD NetCDFs')
@click.argument('preard', type=click.Path(exists=True, resolve_path=True))
@click.option('--dest', type=click.Path(file_okay=False, resolve_path=True),
              help='Override config file destination directory')
@cli_options.opt_executor
@options.opt_overwrite
@click.pass_context
def convert(ctx, preard, dest, overwrite, executor):
    """ Convert "pre-ARD" GeoTIFF(s) to ARD data cubes in NetCDF4 format
    """
    from dask.diagnostics import ProgressBar
    from stems.utils import renamed_upon_completion

    from cedar.preard import (ard_netcdf_encoding, find_preard,
                              process_preard, read_metadata)

    # Provide debug info for the executor
    logger = ctx.obj['logger']
    if executor is not None and logger.level == logging.DEBUG:
        from stems.executor import executor_info
        info = executor_info(executor)
        for i in info:
            logger.debug(i)

    # Get configuration and any encoding provided
    cfg = options.fetch_config(ctx)
    ard_cfg = cfg['ard']
    encoding_cfg = ard_cfg.get('encoding', {})

    preard_files = find_preard(preard)
    if len(preard_files) == 0:
        raise FileNotFoundError('Could not find pre-ARD files to process')
    click.echo(f"Found metadata for {len(preard_files)} pre-ARD to convert")

    # Destination directory from config file, or overriden from CLI
    dest_dir_tmpl = dest or ard_cfg['destination']
    dest_dir_tmpl = os.path.expandvars(dest_dir_tmpl)

    for i, (meta, images) in enumerate(preard_files.items()):
        # Read metadata first so we know what is in order
        metadata = read_metadata(meta)

        # Destination can depend on info in metadata - format it
        dest_dir = create_dest_dir(dest_dir_tmpl, metadata)
        dest_ = dest_dir.joinpath(meta.stem + '.nc')

        if dest_.exists() and not overwrite:
            click.echo(f'Already processed "{meta.stem}" to "{dest_}"')
            continue

        click.echo(f'Processing pre-ARD "{meta.stem}" to destination "{dest_}"')
        dest_.parent.mkdir(parents=True, exist_ok=True)

        # Read TIFF files into ARD-like xr.Dataset
        ard_ds = process_preard(metadata, images)

        # Determine encoding
        encoding = ard_netcdf_encoding(ard_ds, metadata, **encoding_cfg)

        with renamed_upon_completion(dest_) as tmp:
            ard_ds_ = ard_ds.to_netcdf(tmp, encoding=encoding, compute=False)

            # Write with progressbar
            with ProgressBar():
                out = ard_ds_.compute()

    click.echo('Complete')


def create_dest_dir(dest_dir_template, metadata):
    """ Create str format metadata and return formatted template
    """
    from stems.gis.grids import Tile
    namespace = metadata['order'].copy()
    namespace['tile'] = Tile.from_dict(metadata['tile'])
    dest_dir = Path(dest_dir_template.format(**namespace))
    return dest_dir



""" Core functions for metadata creation
"""
import datetime as dt

from .. import __version__
from ..utils import load_ee, serialize_filter

ee = load_ee(False)


def get_program_metadata():
    """ Return metadata about this program (name & version)

    Returns
    -------
    dict
        Program "name" and "version"
    """
    return {
        "name": __package__,
        "version": __version__,
        "ee": ee.__version__
    }


def get_order_metadata(collection, date_start, date_end, filters):
    """ Get metadata about a pre-ARD order against a collection
    """
    return {
        'collection': collection,
        'date_start': date_start.isoformat(),
        'date_end': date_end.isoformat(),
        'filters': [serialize_filter(filter) for filter in filters],
        'submitted': dt.datetime.now().isoformat()
    }


def get_tracking_metadata(tracking_name, tracking_prefix,
                          name_template, prefix_template,
                          collections, tiles):
    """ Get general tracking information about an order
    """
    meta = {
        "submitted": dt.datetime.now().isoformat(),
        "name": tracking_name,
        "prefix": tracking_prefix,
        "collections": list(collections),
        "tiles": [tile.to_dict() for tile in tiles],
        "name_template": name_template,
        "prefix_template": prefix_template
    }
    return meta


def get_submission_info(tile_grid, collections, tile_indices,
                        period_start, period_end, period_freq):
    """ Return information about tracked order submissions
    """
    return {
        'submitted': dt.datetime.today().isoformat(),
        'collections': collections,
        'tile_grid': tile_grid.to_dict(),
        'tile_indices': list(tile_indices),
        'period_start': period_start.isoformat(),
        'period_end': period_end.isoformat(),
        'period_freq': period_freq
    }


def get_task_metadata(task):
    """ Get metadata about an EE task's status

    Parameters
    ----------
    task : ee.batch.task.Task
        Earth Engine task

    Returns
    -------
    dict
        Task metadata
    """
    info = {}

    # When active, task config information has:
    # GDrive keys:
    #    description, dimensions, crs, fileFormat, driveFolder, crs_transform,
    #    driveFileNamePrefix, json
    # GCS keys:
    #    description, dimensions, crs, fileFormat, crs_transform, outputBucket,
    #    outputPrefix, json
    bucket = task.config.get('outputBucket', '')
    if bucket:  # GCS
        info['name'] = task.config['description']
        info['prefix'] = task.config['outputPrefix'].rstrip(name)
    elif 'driveFolder' in task.config:  # GDrive
        info['name'] = task.config['driveFileNamePrefix']
        info['prefix'] = task.config['driveFolder']

    # When submitted/complete, we can get status info
    status = task.status()
    info['status'] = {
        "id": status['id'],
        "state": status["state"],
        # Attributes available post-run
        'creation_timestamp_ms': status.get('creation_timestamp_ms', ''),
        'update_timestamp_ms': status.get('update_timestamp_ms', ''),
        'start_timestamp_ms': status.get('start_timestamp_ms', ''),
        'output_url': status.get('output_url', [])
    }
    return info



import json

import pytest


# import vcrpy
# https://vcrpy.readthedocs.io/en/latest/advanced.html#filter-sensitive-data-from-the-request
# https://vcrpy.readthedocs.io/en/latest/advanced.html#automatic-cassette-naming



""" CLI tools for working with cedar configuration files
"""
import json
import logging
import os
from pathlib import Path
import shutil

import click

from . import options


@click.group('config', help='Create or check cedar configuration files')
@click.pass_context
def group_config(ctx):
    pass


@group_config.command('template', short_help='Generate a config file template')
@click.argument('dest', required=False, type=click.Path(dir_okay=False))
@click.option('--comment', is_flag=True, help='Comment out the template file')
@click.pass_context
def config_template(ctx, dest, comment):
    """ Generate a template configuration file
    """
    from cedar.config import TEMPLATE_FILE

    with open(str(TEMPLATE_FILE)) as template:
        lines = list(template)

    if comment:
        lines = ['# ' + line if not line.startswith('#') else line
                 for line in lines]

    if dest:
        dest = Path(dest)
        dest.parent.mkdir(parents=True, exist_ok=True)

        tmp = f'{dest}.tmp.{os.getpid()}'
        with open(tmp, 'w') as tmp_dst:
            tmp_dst.write(''.join(lines))
        shutil.move(str(tmp), str(dest))

        click.echo(f'Wrote template file to "{dest}"')
    else:
        for line in lines:
            click.echo(line, nl=False)


@group_config.command('build', short_help='Interactively build a config file')
@click.argument('dest', type=click.Path(dir_okay=False))
@click.pass_context
def config_build(ctx, dest):
    click.echo('To do... maybe you?')
    raise click.Abort()


@group_config.command('print', short_help='Parse and print config file')
@click.argument('config_file',
                type=click.Path(dir_okay=False, resolve_path=True))
@click.pass_context
def config_print(ctx, config_file):
    """ This program should help you check your config file is valid
    """
    from jsonschema.exceptions import ValidationError
    from cedar.config import Config

    try:
        cfg = Config.from_yaml(config_file)
    except ValidationError as ve:
        click.echo(f'Configuration file failed to validate: "{ve.message}"')
        raise click.Abort()
    except Exception as e:
        click.echo(f'Could not parse configuration file "{config_file}":\n{e}')
        raise click.Abort()
    else:
        click.echo(cfg.to_yaml(indent=2))



""" Build objects from configuration files
"""
import logging
import os
from pathlib import Path
import yaml

from .. import defaults

logger = logging.getLogger(__name__)


def build_tile_grid(grid_name=None, grid_filename=None, **kwds):
    """ Build the ``tile_grid`` section

    Parameters
    ----------
    grid_name : str
        Name of the grid to use. If using a custom grid, specify the parameters
        for it either by passing the filename of a grid definition file, or by
        passing all the necessary keywords to create a
        :py:class:`stems.gis.grids.TileGrid`.
    grid_filename : str or pathlib.Path, optional
        Custom tile grid definition file. Should be loadable using
        :py:func:`stems.gis.grids.load_grids`.
    kwds
        Otherwise, pass keywords to initialize a TileGrid

    Returns
    -------
    dict
        The ``tile_grid`` parameter section

    See Also
    --------
    stems.gis.grids.TileGrid
    """
    from stems.gis.grids import TileGrid, load_grids

    # Check if user wants a "named" or "defined" grid
    if grid_name:
        # Check for definitions file
        grids = load_grids(filename=filename)
        try:
            tile_grid = grids[grid_name]
        except KeyError:
            known = ', '.join([f'"{k}"' for k in grids])
            raise KeyError(f'Could not load a grid named "{grid_name}". '
                           f'Available grids: {known}')
    else:
        # Should have all the parameters to init otherwise
        tile_grid = TileGrid(**kwds)

    return tile_grid


def build_tracker(tile_grid, store, **kwds):
    """ Build the ``gee`` section and return a Tracker
    """
    from ..tracker import Tracker
    tracker = Tracker(tile_grid, store, **kwds)
    return tracker



""" Helper utilities for using Google Drive
"""
import functools
import io
import json
import logging
import os
from pathlib import Path
import socket
import urllib
from urllib.parse import urlencode
from urllib.error import HTTPError

from apiclient.http import MediaIoBaseDownload
from google.auth.transport.requests import Request
from google.api_core import retry
from google.oauth2.credentials import Credentials
from googleapiclient.discovery import Resource, build
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.http import MediaIoBaseUpload

import ee

from stems.utils import renamed_upon_completion

from .. import defaults, utils, __version__

logger = logging.getLogger(__name__)


MIME_TYPE_DIRECTORY = 'application/vnd.google-apps.folder'
MIME_TYPE_FILE = 'application/vnd.google-apps.file'

_ORDER_BY = 'folder,modifiedTime,name'

METADATA_ENCODING = 'utf-8'

SCOPES = ['https://www.googleapis.com/auth/drive']

# TODO: rehash what these should be and document
_CLIENT_SECRETS = 'client_secrets.json'
_USER_CREDS = 'credentials.json'


def build_gdrive_service(credentials):
    """ Return Google Drive API service, either by credentials or file

    Parameters
    ----------
    credentials : google.oauth2.credentials.Credentials
        User credentials, including access token, for using the application.

    Returns
    -------
    googleapiclient.discovery.Resource
        GDrive v3 API resource
    """
    service = build('drive', 'v3', credentials=credentials)
    return service


def get_credentials(client_secrets_file=None, credentials_file=None,
                    no_browser=True):
    """ Get OAuth2 Credentials for Google Drive

    Parameters
    ----------
    client_secrets_file : str or Path
        Filename of "client_secrets.[...].json" file
    credentials_file : str or Path
        Filename of user credentials to load, or to save to for future use.
        If not provided, will use default location.
    no_browser : bool, optional
        Disables opening a web browser in favor of prompting user to
        authenticate using a terminal prompt.

    Returns
    -------
    credentials : google.oauth2.credentials.Credentials
        User credentials, including access token, for using the application.
    credentials_file : str
        File storing user credentials
    """
    # Locate authentication files, checking default places
    client_secrets_file_, credentials_file_ = find_credentials(
        client_secrets_file, credentials_file)

    creds = None
    if credentials_file_ and os.path.exists(credentials_file_):
        logger.debug('Trying to load previous credentials file...')
        creds = load_credentials(credentials_file_)

    if not creds or not creds.valid:
        # Try refreshing
        if creds and not creds.expired and creds.refresh_token:
            logger.debug('Trying to refresh credentials token...')
            creds.refresh(Request())
        else:
            # Load from file
            if not client_secrets_file_:
                raise ValueError('Missing `client_secrets` or path to a '
                                 'valid `client_secrets_file` file needed '
                                 'to authenticate user')

            logger.debug('Opening local web server to authenticate..')
            flow = InstalledAppFlow.from_client_secrets_file(
                client_secrets_file_, SCOPES)

            flow_kwds = {
                # Enable offline access so we can refresh access token without
                # reprompting user for permission.
                'access_type': 'offline'
            }
            if no_browser:
                creds = flow.run_console(**flow_kwds)
            else:
                creds = flow.run_local_server(**flow_kwds)

        # Save for next time
        logger.debug(f'Saving Google API credentials to {credentials_file_}')
        save_credentials(creds, credentials_file_)

    return creds, credentials_file_


class GDriveStore(object):
    """ Store GEE "pre-ARD" images and metadata on Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google Drive API service
    """
    def __init__(self, service):
        assert isinstance(service, Resource)
        self.service = service

    @classmethod
    def from_credentials(cls, client_secrets_file=None, credentials_file=None):
        """ Create and/or load credentials and create the store

        Parameters
        ----------
        client_secrets_file : str or Path
            Filename of "client_secrets.[...].json" file
        credentials_file : str or Path
            Filename of user credentials to load, or to save to for future use.
            If not provided, will use default location.
        """
        creds, creds_file = get_credentials(
            client_secrets_file=client_secrets_file,
            credentials_file=credentials_file)
        gdrive = build_gdrive_service(credentials=creds)
        return cls(gdrive)

    def list(self, path=None, pattern=None):
        """ List stored images or metadata

        Parameters
        ----------
        path : str, optional
            Prefix path to search within
        pattern : str, optional
            Filename pattern

        Returns
        -------
        list[str]
            Names of stored data
        """
        parent_id = _path_to_parent_id(self.service, path)
        info = list_objects(self.service, parent_id=parent_id, name=pattern)
        return [i['name'] for i in info]

    def store_metadata(self, metadata, name, path=None):
        """ Store JSON metadata

        Parameters
        ----------
        metadata : dict
            Metadata, to be saved as JSON
        name : str
            Name of file/object to store
        path : str, optional
            Parent directory for file/object stored

        Returns
        -------
        str
            ID of file uploaded
        """
        if not name.endswith('.json'):
            name += '.json'

        if path is not None:
            parent_id = mkdir(self.service, path, check=True)

        meta_id = upload_json(self.service, metadata, name,
                              path=path, check=True)
        return meta_id

    def store_image(self, image, name, path=None, **export_image_kwds):
        """ Create ee.batch.Task to create and store "pre-ARD"

        Parameters
        ----------
        image : ee.Image
            Earth Engine image to compute & store
        name : str
            Name of file/object to store
        path : str, optional
            Parent directory for file/object stored
        export_image_kwds : dict, optional
            Additional keyword arguments to pass onto
            :py:meth:`ee.batch.Export.image.toCloudStorage`
            (hint: ``scale`` & ``crs``)

        Returns
        -------
        ee.Task
            Earth Engine Task
        """
        # TODO: warn / fail if `path` looks nested...
        # Make parent directory -- currently GEE interprets this directory as
        # a single, non-nested directory. So, "PREARD/LT05" and "PREARD/LE07"
        # won't be separate directories under "PREARD/", but two directories
        # stored in the root of your drive.
        # As such, we use mkdir here (for now?)
        if path is not None:
            parent_id = mkdir(self.service,  path, check=True)

        # Canonicalized:
        #   folder -> driveFolder
        #   fileNamePrefix, path -> driveFileNamePrefix
        task = ee.batch.Export.image.toDrive(
            image,
            description=name,
            fileNamePrefix=name,
            driveFolder=path,
            **export_image_kwds
        )
        return task

    def _retrieve_extension(self, dest, name, ext, path=None, overwrite=True):
        # Find parent ID if provided & list objects
        parent_id = _path_to_parent_id(self.service, path)
        query = list(list_objects(self.service,
                                  parent_id=parent_id,
                                  name=name))

        # Can't pattern search on GDrive, so limit by extension
        query_ = [q for q in query if q['name'].endswith(ext)]

        for i, result in enumerate(query_):
            id_, result_name = result['id'], result['name']
            msg = f'{i}/{len(query_)} - "{result_name}"'

            dest_ = Path(dest).joinpath(result_name)
            if not dest_.exists() or overwrite:
                logger.debug(f'Downloading {msg}')
                dst = download_file_id(self.service, id_, dest_)
            else:
                logger.debug(f'Already downloaded {msg}')

            yield dest_

    def retrieve_image(self, dest, name, path=None, overwrite=True):
        """ Retrieve (pieces of) an image from the Google Drive

        Parameters
        ----------
        dest : str
            Destination folder to save image(s)
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored on Google Drive

        Yields
        ------
        Sequence[str]
            Filename(s) corresponding to retrieved data
        """
        return self._retrieve_extension(dest, name, '.tif', path=path,
                                        overwrite=overwrite)

    def retrieve_metadata(self, dest, name, path=None, overwrite=True):
        """ Retrieve image metadata from the GCS

        Parameters
        ----------
        dest : str
            Destination folder to save metadata
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored

        Yields
        ------
        pathlib.Path
            Filename corresponding to retrieved data
        """
        return self._retrieve_extension(dest, name, '.json', path=path,
                                        overwrite=overwrite)

    def read_metadata(self, name, path=None):
        """ Read and parse JSON metadata into a dict

        Parameters
        ----------
        name : str
            Filename of metadata to read
        path : str, optional
            Parent directory for file/object stored

        Returns
        -------
        dict
            JSON metadata
        """
        parent_id = _path_to_parent_id(self.service, path)
        return read_json(self.service, name, parent_id=parent_id)

    def remove(self, name, path=None):
        """ Remove a file from Google Drive

        Parameters
        ----------
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored

        Returns
        -------
        str
            Name of file removed
        """
        parent_id = _path_to_parent_id(self.service, path)
        return delete(self.service, name, parent_id=parent_id)


def get_appProperties():
    """ Returns private "appProperties" to include when creating files
    """
    return {
        'application': __name__.split('.')[0],
        'version': __version__
    }


def query_appProperties():
    """ Build a query component for ``appProperties``
    """
    q = ('appProperties has {key="application" and value="%s"}'
         % get_appProperties()['application'])
    return q


def upload_json(service, data, name, path=None, check=False,
                encoding=METADATA_ENCODING):
    """ Upload JSON data to Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    data : str or dict
        JSON, either already dumped to str or as a dict
    name : str
        Filename for data
    path : str, optional
        Parent directory to put file. If ``None``, stores in
        root of Google Drive
    check : bool, optional
        Check to see if the file already exists first. If so, will
        overwrite (or "update" instead of "create")
    encoding : str, optional
        Metadata encoding

    Returns
    -------
    str
        ID of file uploaded
    """
    # Dump to JSON if needed
    if isinstance(data, dict):
        data = json.dumps(data, indent=2)
    if not isinstance(data, bytes):
        data = data.encode(encoding)

    # Find folder ID for parent directory
    parent_id = _path_to_parent_id(service, path)

    # Prepare metadata body & data
    body = {
        'name': name,
        'parents': [parent_id],
        'mimeType': 'text/plain',
        'appProperties': get_appProperties()
    }
    content = io.BytesIO(data)
    media = MediaIoBaseUpload(content, 'text/plain', resumable=True)

    # Check to see if file already exists...
    if check:
        file_id = exists(service, name, parent_id=parent_id)
    else:
        file_id = ''

    # Update or create
    if check and file_id:
        req = service.files().update(fileId=file_id, media_body=media)
    else:
        req = service.files().create(body=body, media_body=media, fields='id')
    meta = req.execute()

    return meta['id']


@retry.Retry()
def download_file_id(service, file_id, dest):
    """ Download a file to a destination directory using its ID

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    file_id : str
        ID of file on Google Drive
    dest : str or pathlib.Path
        Destination filename

    Returns
    -------
    pathlib.Path
        Filename written to

    Raises
    ------
    FileExistsError
        Raised if file exists in destination but not allowed to overwrite,
    ValueError
        Raised if the file given does not exist in Google Drive
    """
    # Create destination if needed
    dest = Path(dest)
    if not dest.parent.exists():
        dest.parent.mkdir(parents=True, exist_ok=True)

    # Download...
    request = service.files().get_media(fileId=file_id)

    suffix = f'.tmp.{socket.gethostname()}{os.getpid()}'
    with renamed_upon_completion(dest, suffix=suffix) as tmp:
        with open(str(tmp), 'wb') as dst:
            downloader = MediaIoBaseDownload(dst, request)
            done = False
            # TODO: would be nice to have _some_ logging indicator of dl speed
            while done is False:
                status, done = downloader.next_chunk()

    return dest


def download_file(service, name, dest, parent_id=None):
    """ Download a file to a destination directory

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    dest : str
        Name of destination
    parent_id : str, optional
        Parent ID of folder containing file (to narrow search)

    Returns
    -------
    pathlib.Path
        Filename written to

    Raises
    ------
    FileExistsError
        Raised if file exists in destination but not allowed to overwrite,
    ValueError
        Raised if the file given does not exist in Google Drive
    """
    # Find the file
    name_id = exists(service, name, parent_id=parent_id)
    if not name_id:
        raise ValueError(f'File "{name}" not found on Google Drive')

    return download_file_id(service, name_id, dest)


@retry.Retry()
def _read_json_id(service, file_id):
    request = service.files().get_media(fileId=file_id)
    with io.BytesIO() as fh:
        downloader = MediaIoBaseDownload(fh, request)
        done = False
        while done is False:
            status, done = downloader.next_chunk()
        fh.seek(0)
        data = fh.read()
        return json.loads(data, encoding=METADATA_ENCODING)


def read_json(service, name, parent_id=None):
    """ Reads and returns a JSON file from Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    name : str
        Name of file/folder
    parent_id : str, optional
        Parent ID of folder containing file (to narrow search)

    Returns
    -------
    dict
        JSON data as a dict
    """
    name_id = exists(service, name, parent_id=parent_id)
    if not name_id:
        raise ValueError(f'File "{name}" not found on Google Drive')
    return _read_json_id(service, name_id)


def mkdir_p(service, dest, parent_id=None):
    """ Make a directory, recursively

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    dest : str
        Directory to create

    Returns
    -------
    str
        Google Drive ID for directory created (or already existing)
    """
    paths = dest.split('/', 1)
    if len(paths) == 1:  # root
        name_id = exists(service, paths[0], parent_id=parent_id)
        if not name_id:
            return mkdir(service, paths[0], parent_id=parent_id)
        else:  # already exists
            return name_id
    else:
        name, paths_ = paths[0], paths[1:]
        name_id = exists(service, name, parent_id=parent_id, directory=True)
        if not name_id:
            name_id = mkdir(service, name, parent_id=parent_id)
        dest_ = '/'.join(paths_)
        return mkdir_p(service, dest_, parent_id=name_id)


@retry.Retry()
def mkdir(service, name, parent_id=None, check=False):
    """ Make a directory on GDrive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    name : str
        Directory to create
    parent_id : str, optional
        ID of parent directory
    check : bool, optional
        Check if directory exists before creating. If exists, will not
        create a new directory and instead return the ID of this directory

    Returns
    -------
    str
        Google Drive ID for directory created (or already existing)
    """
    if check:
        name_id = exists(service, name, parent_id=parent_id, directory=True)
        if name_id:
            logger.debug('Not creating new directory; already exists')
            return name_id

    logger.debug(f'Creating directory {name}')
    meta = {'name': name, 'mimeType': MIME_TYPE_DIRECTORY}
    if parent_id is not None:
        meta['parents'] = [parent_id]

    dir_ = service.files().create(body=meta, fields='id').execute()
    return dir_['id']


# TODO: maybe also make this have a timeout for memoized keys
def _memoize_exists(func):
    cache = {}
    @functools.wraps(func)
    def inner(*args, **kwds):
        # don't use "service" as key
        key = (args[1],
               kwds.get('parent_id', None),
               kwds.get('directory', False),
               kwds.get('trashed', False), )
        if key in cache:
            return cache[key]
        else:
            ans = func(*args, **kwds)
            # only cache "hits" (id != '')
            if ans:
                cache[key] = ans
            return ans

    return inner


@retry.Retry()
@_memoize_exists
def exists(service, name, parent_id=None, directory=False, trashed=False,
           appProperties=defaults.GDRIVE_USE_APPPROPERTIES):
    """ Check if file/folder exists

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    name : str
        Name of file/folder
    parent_id : str, optional
        Parent ID of folder containing file (to narrow search)
    directory : bool, optional
        True if file needs to also be a directory
    trashed : bool, optional
        Search in the trash?
    appProperties : bool
        Search for application-specific files using ``appProperties``

    Returns
    -------
    str
        Returns object ID if exists, otherwise empty string
    """
    q = []
    if directory:
        q.append(f'mimeType = "{MIME_TYPE_DIRECTORY}"')
    if appProperties:
        q.append(query_appProperties())

    results = list(list_objects(service, parent_id=parent_id, name=name, q=q,
                                appProperties=appProperties))
    n_results = len(results)

    if n_results == 1:
        return results[0]['id']
    elif n_results > 1:
        match_name = [r['name'] == name for r in results]
        if any(match_name):
            idx = match_name.index(True)
            return results[idx]['id']
        else:
            id_ = results[0]['id']
            logger.debug(f'Returning first of {n_results} matches -- id={id_}')
            return id_
    else:
        return ''


# TODO: memoize
@retry.Retry()
def list_objects(service, parent_id=None, name=None, q=None,
                 appProperties=defaults.GDRIVE_USE_APPPROPERTIES):
    """ List files/folders on Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    parent_id : str, optional
        Parent ID of folder to list (to narrow search)
    name : str, optional
        Name to search for (don't include asterisks)
    q : str or Sequence[str], optional
        Additional search query parameters
    appProperties : bool
        Search for application-specific files using ``appProperties``

    Yields
    ------
    list[dict]
        Info about objects stored on Google Drive (keys=('id', 'name', ))
    """
    # Check for passed query
    if isinstance(q, str):
        query = [q]
    elif isinstance(q, (list, tuple)):
        query = list(q)
    else:
        query = []

    # Build up query
    query.append("trashed = false")
    if parent_id:
        query.append(f"'{parent_id}' in parents")
    if name:
        query.append(f"name contains '{name}'")
    if appProperties:
        query.append(query_appProperties())

    # Combine
    query_ = ' and '.join(query)
    logger.debug(f'Searching for query: "{query_}"')

    # Execute in pages
    page_token = None
    while True:
        resp = service.files().list(q=query_,
                                    spaces='drive',
                                    orderBy=_ORDER_BY,
                                    fields='nextPageToken, files(id, name)',
                                    pageToken=page_token).execute()
        for file_ in resp.get('files', []):
            yield file_
        page_token = resp.get('nextPageToken', None)
        if page_token is None:
            break


def list_dirs(service, parent_id=None, name=None):
    """ List folders on Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    parent_id : str, optional
        Parent ID of folder to list (to narrow search)
    name : str, optional
        Name to search for (don't include asterisks)

    Yields
    ------
    list[dict]
        Info about objects stored on Google Drive (keys=('id', 'name', ))
    """
    q = [f'mimeType = "{MIME_TYPE_DIRECTORY}"']
    return list_objects(service, parent_id=parent_id, name=name, q=q)


@retry.Retry()
def delete(service, name, parent_id=None,
           appProperties=defaults.GDRIVE_USE_APPPROPERTIES):
    """ Delete a file/folder on Google Drive

    Parameters
    ----------
    service : googleapiclient.discovery.Resource
        Google API resource for GDrive v3
    name : str
        Name of file/folder
    parent_id : str, optional
        Parent ID of folder containing file (to narrow search)
    appProperties : bool
        Search for application-specific files using ``appProperties``

    Returns
    -------
    str
        ID of deleted file/folder
    """
    name_id = exists(service, name, parent_id=parent_id)
    resp = service.files().delete(fileId=name_id).execute()
    return name_id


def _path_to_parent_id(service, path):
    if path is None:
        return None
    else:
        # Never search using appProperties
        # (e.g., in case GEE created the export folder)
        parent_id = exists(service, path, directory=True,
                           appProperties=False)
        if not parent_id:
            raise FileNotFoundError(
                f'Cannot find prefix path provided "{path}" on Google Drive')
        return parent_id


# =============================================================================
# OAuth helpers
_OAUTH2_CREDS = ['token', 'refresh_token', 'id_token', 'token_uri',
                 'client_id', 'client_secret', 'scopes']


def _dict_to_creds(d):
    assert isinstance(d, dict)
    token = d['token']
    kwds = {k: v for k, v in d.items() if k != 'token'}
    creds = Credentials(token, **kwds)
    return creds


def _creds_to_dict(creds):
    creds_ = {
        k: getattr(creds, k, None) for k in _OAUTH2_CREDS
    }
    return creds_


def load_credentials(filename):
    """ Load Google OAuth2 credentials from file

    Returns
    -------
    google.oauth2.credentials.Credentials
    """
    with open(filename) as src:
        data = json.load(src)
    creds = _dict_to_creds(data)
    return creds


def save_credentials(credentials, filename):
    """ Save Google OAuth2 credentials

    Parameters
    ----------
    credentials : google.oauth2.credentials.Credentials
        Google OAuth2 credentials
    filename : str or Path
        Path to save credentials
    """
    creds_ = _creds_to_dict(credentials)

    filename = Path(filename)
    filename.parent.mkdir(exist_ok=True, parents=True)

    suffix = f'.tmp.{socket.gethostname()}{os.getpid()}'
    with renamed_upon_completion(filename, suffix=suffix) as tmp:
        with open(str(tmp), 'w') as dst:
            json.dump(creds_, dst, indent=2, sort_keys=False)
        utils.set_file_urw(tmp)


def find_credentials(client_secrets_file=None, credentials_file=None):
    """ Locate GDrive client secrets & credentials files
    """
    def _check(fname, exists=True):
        dname, bname = os.path.dirname(fname), os.path.basename(fname)
        dirs_ = defaults.CEDAR_ROOT_CONFIG.copy()
        if dname:
            dirs_.insert(0, dname)
        paths = [os.path.join(d, bname) for d in dirs_]
        return utils.get_file(*paths, exists=exists)

    client_secrets_file_ = _check(client_secrets_file or _CLIENT_SECRETS, True)
    credentials_file_ = _check(credentials_file or _USER_CREDS, False)

    return client_secrets_file_, credentials_file_



""" JSON Schema validation helpers
"""
import json
import os
from pathlib import Path

from jsonschema import Draft7Validator, RefResolver, validators


def validate_with_defaults(obj, schema, resolve=None):
    """ Validate a configuration file and fill in defaults (modifies ``obj``)
    """
    if resolve:
        resolver = RefResolver('file://' + str(resolve) + '/', schema)
    else:
        resolver = None

    # Allow tuple/list as 'array' type
    # See: https://github.com/Julian/jsonschema/issues/148
    types_ = {'array': (list, tuple)}

    validator = DefaultValidatingDraft7Validator(schema, resolver=resolver,
                                                 types=types_)
    validator.validate(obj)


# Adapted from python-jsonschema FAQ
# https://python-jsonschema.readthedocs.io/en/stable/faq/
def _extend_with_default(validator_class):
    validate_properties = validator_class.VALIDATORS["properties"]

    def set_defaults(validator, properties, instance, schema):
        for property, subschema in properties.items():
            if "default" in subschema:
                instance.setdefault(property, subschema["default"])

        for error in validate_properties(
                validator, properties, instance, schema):
            yield error

    return validators.extend(validator_class, {"properties" : set_defaults})


DefaultValidatingDraft7Validator = _extend_with_default(Draft7Validator)



""" CLI for downloading data
"""
from pathlib import Path

import click

from . import options


@click.command('download',
               short_help='Download exported "pre-ARD" data from storage')
@options.arg_tracking_name
@options.opt_update_order
@click.option('--clean', 'clean_', is_flag=True,
              help=('Run `cedar clean` for order if completed and '
                    'successfully downloaded'))
@click.option('--dest', 'dest_dir',
              type=click.Path(file_okay=False, resolve_path=True),
              help='Specify destination root directory. If not specified, '
                   'downloads order into current directory')
@options.opt_overwrite
@click.pass_context
def download(ctx, tracking_name, update_, clean_, dest_dir, overwrite):
    """ Download pre-ARD for a tracked order

    Downloads data into a directory named after the order name. Pass
    ``--dest`` to change the root location of this directory.

    \b
    TODO
    ----
    * Silence / don't use progressbar if we're quiet
    """
    from cedar.utils import load_ee

    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    # Rename to remove .json
    tracking_name = tracking_name.rstrip('.json')

    # Destination defaults to tracking_name
    if dest_dir is None:
        # remove any extension listed
        dest_dir = Path('.').resolve()
    else:
        dest_dir = Path(dest_dir)
    dest = dest_dir.joinpath(tracking_name)

    click.echo(f'Retrieving pre-ARD from tracking info: {tracking_name}')
    if update_:
        load_ee(True)  # authenticate against EE API
        tracking_info = tracker.update(tracking_name)
    else:
        tracking_info = tracker.read(tracking_name)

    n_tasks = len(tracking_info['orders'])
    click.echo(f'Downloading data for {n_tasks} tasks')
    with click.progressbar(label='Downloading',
                           item_show_func=_item_show_func,
                           length=n_tasks) as bar:
        cb_bar = _make_callback(bar)
        dl_info = tracker.download(tracking_info, dest,
                                   overwrite=overwrite, callback=cb_bar)

    if clean_:
        if tracking_info.complete:
            clean_info = _do_clean(tracker, tracking_name, tracking_info, False)
        else:
            click.echo('Not cleaning data because the order is still in '
                       'progress or the tracking metadata has not been '
                       f'updated. Run ``cedar status update {tracking_name}`` '
                       'if you believe the order has completed')

    click.echo('Complete!')


@click.command('clean',
               short_help='Clean/delete exported "pre-ARD" data from storage')
@click.option('--keep-tracking', is_flag=True,
              help='Preserve tracking information (deleted by default)')
@options.arg_tracking_name
@click.pass_context
def clean(ctx, tracking_name, keep_tracking):
    """ Clean pre-ARD described by tracking information
    """
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    click.echo(f'Retrieving info about pre-ARD in "{tracking_name}"')
    tracking_info = tracker.read(tracking_name)

    click.echo(f'Cleaning data for {n_orders} orders')
    clean_info = _do_clean(tracker, tracking_name, tracking_info, keep_tracking)

    click.echo('Complete!')


def _do_clean(tracker, tracking_name, tracking_info, keep_tracking=False):
    n_orders = len(tracking_info['orders'])
    with click.progressbar(label='Cleaning',
                           item_show_func=_item_show_func,
                           length=n_orders) as bar:
        cb_bar = _make_callback(bar)
        clean_info = tracker.clean(
            tracking_info,
            tracking_name=None if keep_tracking else tracking_name,
            callback=cb_bar
        )
    return clean_info


def _make_callback(bar):
    def callback(item, n_steps):
        bar.current_item = item
        bar.update(n_steps)
    return callback


def _item_show_func(item):
    return str(item) if item else ''



""" Configuration file handling
"""
from collections import Mapping
import logging
import json
import os
from pathlib import Path
import yaml

from .. import defaults, validation
from . import build

logger = logging.getLogger(__name__)


TEMPLATE_FILENAME = 'config.yaml.tmpl'
TEMPLATE_FILE = Path(__file__).parent.joinpath(TEMPLATE_FILENAME)

SCHEMA_FILENAME = 'schema.json'
SCHEMA_FILE = os.path.join(os.path.dirname(__file__), SCHEMA_FILENAME)


class Config(Mapping):
    """ CEDAR configuration file
    """

    def __init__(self, config, schema=None):
        self._config = config.copy()
        self.schema = schema or self._load_schema()
        self.validate()

    # Mapping methods
    def __getitem__(self, key):
        return self._config[key]

    def __iter__(self):
        for key in self._config:
            yield key

    def __len__(self):
        return len(self._config)

    # Class create/serialize methods
    @classmethod
    def from_yaml(cls, filename, schema=None):
        """ Load from a YAML configuration file
        """
        with open(filename) as f:
            config = yaml.safe_load(f)
        return cls(config, schema=schema)

    @classmethod
    def from_template(cls, schema=None):
        """ Load from the included YAML configuration file template
        """
        return cls.from_yaml(TEMPLATE_FILE, schema=schema)

    def to_yaml(self, dest=None, indent=2, sort_keys=False, **kwds):
        """ Write to a YAML (file, if ``dest`` is provided)

        Parameters
        ----------
        dest : str or Path
            Filename to write to. If not provided, returns a str

        Returns
        -------
        str
            Either the filename written to, or a str containing the
            YAML data if ``dest`` is ``None``
        """
        if dest is not None:
            with open(dest, 'w') as dst:
                dmp = yaml.safe_dump(self._config, stream=dst,
                                     indent=indent, sort_keys=sort_keys,
                                     **kwds)
            return dst
        else:
            return yaml.safe_dump(self._config,
                                  indent=indent, sort_keys=sort_keys,
                                  **kwds)

    # Getter-s for various objects this config describes
    @staticmethod
    def _load_schema(filename=SCHEMA_FILE):
        with open(filename) as src:
            return json.load(src)

    def validate(self):
        """ Validate the configuration against schema

        Raises
        ------
        ValidationError
            Raised if there's an issue
        """
        validation.validate_with_defaults(self._config, schema=self.schema)

    def get_tracker(self):
        """ Get the Tracker described by this store
        """
        # Copy tracker config
        cfg = self['tracker'].copy()

        # Create tile grid
        tile_grid = self.get_tile_grid()

        # Create store
        service = cfg.pop('store').lower()
        if service == 'gcs':
            store = self.get_gcs_store()
        elif service == 'gdrive':
            store = self.get_gdrive_store()
        else:
            raise ValueError(f'Unknown `store_service` named "{service}"')

        # Create tracker
        tracker = build.build_tracker(tile_grid, store, **cfg)
        return tracker

    def get_tile_grid(self):
        """ Return the Tile Grid described by this config

        Returns
        -------
        stems.gis.grids.TileGrid
        """
        cfg = self['tile_grid'].copy()
        grid_name = cfg.pop('grid_name', None)
        grid_filename = cfg.pop('grid_filename', None)
        return build.build_tile_grid(grid_name, grid_filename, **cfg)

    def get_gcs_store(self):
        """ Return a GCSStore described by this config
        """
        from cedar.stores.gcs import GCSStore
        cfg = self.get('gcs', {})
        store = GCS.from_credentials(**kwds)
        return store

    def get_gdrive_store(self):
        """ Return a GDriveStore described by this config
        """
        from cedar.stores.gdrive import GDriveStore
        cfg = self.get('gdrive', {})
        store = GDriveStore.from_credentials(**cfg)
        return store



""" Helper utilities for using Google Cloud Storage
"""
import fnmatch
import functools
import io
import json
import logging
import os
from pathlib import Path
import re

from google.cloud import storage
from google.oauth2 import service_account

import ee

from stems.utils import renamed_upon_completion

logger = logging.getLogger(__name__)

_RE_FILE = re.compile(r'.*(?<!\/)$')
METADATA_ENCODING = 'utf-8'

SCOPES = [
    'https://www.googleapis.com/auth/devstorage.read_write'
]
ENVVAR_CREDENTIALS = 'GOOGLE_APPLICATION_CREDENTIALS'
ENVVAR_PROJECT = 'GOOGLE_CLOUD_PROJECT'


def build_gcs_client(credentials=None, project=None):
    """ Return a Google Cloud Store API service client

    Parameters
    ----------
    credentials : str, optional
        File name of Google Cloud Store credentials (typically from a service
        account)
    project : str, optional
        Google Cloud Platform project to use

    Returns
    -------
    google.cloud.storage.Client
        Client for the Google Cloud Storage client library

    Notes
    -----
    You might consider setting the envirnment variable
    ``GOOGLE_APPLICATION_CREDENTIALS`` with the path to your service account
    credentials file [1]_.

    References
    ----------
    .. [1] https://cloud.google.com/storage/docs/reference/libraries#setting_up_authentication
    """
    if credentials is None and ENVVAR_CREDENTIALS in os.environ:
        credentials = os.environ[ENVVAR_CREDENTIALS]
        logger.debug('Got GCS credentials from environment variable')
    if project is None and ENVVAR_PROJECT in os.environ:
        project = os.environ[ENVVAR_PROJECT]
        logger.debug('Got GCS project from environment variable')

    credentials = service_account.Credentials.from_service_account_file(
        credentials,
        scopes=SCOPES)
    client = storage.Client(project=project, credentials=credentials)

    return client


class GCSStore(object):
    """ Store GEE "pre-ARD" images and metadata on Google Cloud Storage

    Parameters
    ----------
    client : google.cloud.storage.client.Client
        GCS client
    bucket : google.cloud.storage.bucket.Bucket
        GCS bucket
    """
    def __init__(self, client, bucket):
        assert isinstance(client, storage.Client)
        assert isinstance(bucket, storage.Bucket)
        self.client = client
        self.bucket = bucket

    @classmethod
    def from_credentials(cls, bucket_name, credentials=None, project=None):
        """ Load Google Cloud Storage credentials and create store
        """
        client = build_gcs_client(credentials=credentials, project=project)
        bucket = client.get_bucket(bucket_name)
        return cls(client, bucket)

    def list(self, path=None, pattern=None):
        """ List stored images or metadata

        Parameters
        ----------
        path : str, optional
            Prefix path to search within
        pattern : str, optional
            Filename pattern

        Returns
        -------
        list[str]
            Names of stored data
        """
        blobs = list_blobs(self.bucket, prefix=path, pattern=pattern)
        return [blob.name for blob in blobs]

    def store_metadata(self, metadata, name, path=None):
        """ Store JSON metadata

        Parameters
        ----------
        metadata : dict
            Metadata, to be saved as JSON
        name : str
            Name of file/object to store
        path : str, optional
            Parent directory for file/object stored. Otherwise assumed to
            be part of ``name``

        Returns
        -------
        str
            Path to uploaded object
        """
        if not name.endswith('.json'):
            name += '.json'
        fullname, _, path_ = _combine_name_path(name, path)

        path_ = mkdir_p(self.bucket, path_)

        blob = upload_json(self.bucket, metadata, fullname,
                           check=False)
        return blob.name

    def store_image(self, image, name, path=None, **export_image_kwds):
        """ Create ee.batch.Task to create and store "pre-ARD"

        Parameters
        ----------
        image : ee.Image
            Earth Engine image to compute & store
        name : str
            Name of file/object to store
        path : str, optional
            Parent directory for file/object stored
        export_image_kwds : dict, optional
            Additional keyword arguments to pass onto
            :py:meth:`ee.batch.Export.image.toCloudStorage` (hint: ``scale`` & ``crs``)

        Returns
        -------
        ee.Task
            Earth Engine Task
        """
        # Make parent directory
        fullname, basename, path_ = _combine_name_path(name, path)
        path_ = mkdir_p(self.bucket, path_)

        # Create compute/store export task
        # Canonicalized:
        #   bucket -> outputBucket
        #   fileNamePrefix, path -> outputPrefix
        task = ee.batch.Export.image.toCloudStorage(
            image,
            bucket=self.bucket.name,
            description=basename,
            fileNamePrefix=fullname,
            **export_image_kwds
        )
        return task

    def _retrieve_extension(self, dest, name, ext, path=None, overwrite=True):
        if not name.endswith(ext):
            name += f'*{ext}'

        blobs = list_blobs(self.bucket, prefix=path, pattern=name)
        logger.debug(f'Found {len(blobs)} blobs matching image name/prefix')

        dests = []
        for blob in blobs:
            blob_name = blob.name.split('/')[-1]
            msg = f'{i}/{len(blobs)} - "{blob_name}"'

            dest_ = dest.joinpath(blob_name)
            if not dest_.exists() or overwrite:
                logger.debug(f'Downloading {msg}')
                dst_ = download_blob(blob, dest_)
            else:
                logger.debug(f'Already downloaded {msg}')

            yield dest_

    def retrieve_image(self, dest, name, path=None, overwrite=True):
        """ Retrieve (pieces of) an image from the GCS

        Parameters
        ----------
        dest : str
            Destination folder to save image(s)
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored on GCS

        Yields
        ------
        Sequence[str]
            Filename(s) corresponding to retrieved data
        """
        return _retrieve_extension(dest, name, 'tif', path=path,
                                   overwrite=overwrite)

    def retrieve_metadata(self, dest, name, path=None, overwrite=True):
        """ Retrieve image metadata from the GCS

        Parameters
        ----------
        dest : str
            Destination folder to save metadata
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored

        Yields
        ------
        pathlib.Path
            Filename corresponding to retrieved data
        """
        return _retrieve_extension(dest, name, 'json', path=path,
                                   overwrite=overwrite)

    def read_metadata(self, name, path=None):
        """ Read and parse JSON metadata into a dict

        Parameters
        ----------
        name : str
            Filename of metadata blob to read
        path : str, optional
            Parent directory for file/object stored

        Returns
        -------
        dict
            JSON metadata blob
        """
        fullname, _, _ = _combine_name_path(name, path)
        blob = self.bucket.get_blob(fullname)
        if not blob:
            raise ValueError(f'No stored metadata named {fullname}')
        data = read_json(blob, encoding=METADATA_ENCODING)
        return data

    def remove(self, name, path=None):
        """ Remove a file from GCS

        Parameters
        ----------
        name : str
            Name of stored file/object
        path : str, optional
            Parent directory for file/object stored

        Returns
        -------
        str
            Name of file removed
        """
        fullname, _, _ = _combine_name_path(name, path)
        return delete_blob(self.bucket, fullname)


def upload_json(bucket, data, path, check=False, encoding=METADATA_ENCODING):
    """ Upload data as JSON to GCS

    Parameters
    ----------
    bucket : google.cloud.storage.bucket.Bucket
        GCS bucket
    data : str or dict
        JSON, either already dumped to str or as a dict
    path : str
        Destination path on GCS for data
    check : bool, optional
        Check to see if the file already exists first. If so, will
        overwrite (or "update" instead of "create")
    encoding : str, optional
        Metadata encoding

    Returns
    -------
    google.cloud.storage.blob.Blob
        JSON as GCS blob
    """
    blob = bucket.blob(path)
    # Dump dict to JSON str and/or encode if needed
    if isinstance(data, dict):
        data = json.dumps(data, indent=2)
    if not isinstance(data, bytes):
        data = data.encode(encoding)
    blob.upload_from_string(data, content_type='application/json')
    return blob


def read_json(blob, encoding=METADATA_ENCODING):
    """ Read a blob of JSON string data into a dict

    Parameters
    ----------
    blob : google.cloud.storage.blob.Blob
        Blob to read, decode, and load
    encoding : str, optional
        Metadata encoding

    Returns
    -------
    dict
        Blob read, decoded, and loaded as a dict
    """
    data = blob.download_as_string().decode(encoding)
    return json.loads(data)


def mkdir_p(bucket, path):
    """ Create a "directory" on GCS


    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    path : str
        Path to folder

    Returns
    -------
    google.cloud.storage.blob.Blob
        GCS blob for directory created

    Notes
    -----
    Directories don't really exist on GCS but we can fake it [1]_.

    References
    ----------
    .. [1] https://cloud.google.com/storage/docs/gsutil/addlhelp/HowSubdirectoriesWork
    """
    paths = path.rstrip('/').split('/')
    blob_dir = None
    for i in range(len(paths)):
        path_ = _format_dirpath('/'.join(paths[:i + 1]))
        if not exists(bucket, path_):
            logger.debug(f'Creating "directory" on GCS "{path_}"')
            blob_dir = mkdir(bucket, path_)
        else:
            logger.debug(f'Path {path_} already exists...')

    return blob_dir


def mkdir(bucket, path):
    """ Make a directory, recursively

    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    path : str
        Path to folder

    Returns
    -------
    google.cloud.storage.blob.Blob
        GCS blob for directory created
    """
    path_ = _format_dirpath(path)
    blob = bucket.blob(path_)
    content_type = 'application/x-www-form-urlencoded;charset=UTF-8'
    blob.upload_from_string('', content_type=content_type)
    return path


def exists(bucket, path):
    """ Check if file/folder exists

    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    path : str
        Path to file/folder

    Returns
    -------
    bool
        True or False if exists
    """
    blob = storage.Blob(path, bucket)
    return blob.exists()


def delete_blob(bucket, path):
    """ Delete a GCS blob

    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    path : str
        Path to file/folder

    Returns
    -------
    path
        Name of deleted blob
    """
    blob = bucket.blob(path)
    blob.delete()
    return blob.name


def list_dirs(bucket, prefix=None):
    """ Return "directory" blobs within a on GCS

    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    prefix : str
        List contents within this folder

    Returns
    -------
    Sequence[str]
        List of directories names inside at ``prefix``
    """
    prefix = _format_dirpath(prefix) if prefix else None
    blobs = bucket.list_blobs(prefix=prefix, delimiter='/')
    prefixes = set()
    for page in blobs.pages:
        prefixes.update(page.prefixes)
    return list(prefixes)


def list_blobs(bucket, prefix=None, pattern=None):
    """ Return file/non-directory blobs within a on GCS

    Parameters
    ----------
    bucket : str or google.cloud.storage.bucket.Bucket
        Bucket or bucket name
    prefix : str
        List contents within this folder
    pattern : str, optional
        Filter search by glob pattern (i.e., ``*.json``)

    Returns
    -------
    list[google.cloud.storage.blob.Blob]
        List of blob files inside at ``prefix``
    """
    if prefix:
        prefix = _format_dirpath(prefix)

    blobs = bucket.list_blobs(prefix=prefix)
    blobs = [blob for blob in blobs if _RE_FILE.match(blob.name)]

    if pattern:
        re_pattern = re.compile(fnmatch.translate(pattern))
        blobs = [blob for blob in blobs
                 if re_pattern.match(_blob_basename(blob, prefix))]

    return blobs


def download_blob(blob, dest):
    """ Download a blob to a destination directory

    Parameters
    ----------
    blob : google.cloud.storage.blob.Blob
        GCS blob to download
    dest : str
        Local directory to download blob into

    Returns
    -------
    pathlib.Path
        Filename written to
    """
    dest = Path(dest)
    if not dest.exists():
        dest.parent.mkdir(parents=True, exist_ok=True)

    suffix = f'.tmp.{socket.gethostname()}.{os.getpid()}'
    with renamed_upon_completion(dest, suffix=suffix) as tmp:
        blob.download_to_filename(tmp)
    return dest


def _format_dirpath(path):
    return path if path.endswith('/') else path + '/'


def _blob_basename(blob, prefix):
    if prefix and blob.name.startswith(prefix):
        return blob.name[len(prefix):]
    else:
        return blob.name


def _combine_name_path(name, prefix=None):
    # Entire filename
    if prefix is None:
        parts = name.split('/')
        prefix_ = '/'.join(parts[:-1])
        basename = parts[-1]
        return name, basename, _format_dirpath(prefix_)
    # Filename + prefix
    else:
        assert '/' not in name  # should just be a file
        prefix_ = _format_dirpath(prefix)
        return prefix_ + name, name, prefix_



""" Classes and other core tools for tracking & image metadata
"""
from collections import Mapping, defaultdict
import datetime as dt
import json
from pathlib import Path

import numpy as np

from .core import get_task_metadata
from .. import validation


_HERE = Path(__file__).resolve().parent
SCHEMA_TRACKING = _HERE.joinpath('schema_tracking.json')
SCHEMA_IMAGE = _HERE.joinpath('schema_image.json')


class TrackingMetadata(Mapping):
    """ CEDAR order tracking metadata

    Parameters
    ----------
    metadata : dict
        Tracking metadata information as a dict
    schema : dict, optional
        Validate metadata against a specific schema
    """
    def __init__(self, metadata, schema=None):
        self._metadata = metadata
        self.schema = schema or self._load_schema()
        self.validate()

    @classmethod
    def from_file(cls, filename, schema=None):
        """ Create from a tracking metadata file
        """
        with open(filename) as src:
            metadata = json.load(src)
        return cls(metadata, schema=schema)

    @classmethod
    def from_json(cls, json_str, schema=None):
        """ Load tracking metadata from a JSON string
        """
        metadata = json.loads(json_str)
        return cls(metadata, schema=schema)

    # Mapping object requirements
    def __getitem__(self, key):
        return self._metadata[key]

    def __len__(self):
        return len(self._metadata)

    def __iter__(self):
        return iter(self._metadata)

    # Main sections -- make them available as attributes
    @property
    def program(self):
        return self['program']

    @property
    def submission(self):
        return self['submission']

    @property
    def tracking(self):
        return self['tracking']

    @property
    def orders(self):
        return self['orders']

    @property
    def metadata(self):
        return self['metadata']

    @property
    def states(self):
        """ dict: Summary of EE task status for this order
        """
        return summarize_states(self.orders)

    @property
    def progress(self):
        """ float: Percent of EE tasks completed in this order
        """
        states = self.states
        n = sum(states.values())
        n_completed = states.get(_EE_COMPLETED, 0)
        return n_completed / n if n_completed else 0.

    @property
    def complete(self):
        """ bool: True if all orders have completed
        """
        return list(self.states) == [_EE_COMPLETED]

    @property
    def tasks(self):
        """ List[ee.batch.Task]: EarthEngine tasks associated with this order
        """
        from ..utils import get_ee_tasks, load_ee
        ee_api = load_ee(False)  # should initialize elsewhere
        tasks = get_ee_tasks()
        order_tasks = [tasks[order['status']['id']] for order in self.orders]
        return order_tasks

    def update(self):
        """ Update the tracking metadata by checking EE task status

        Returns
        -------
        self
            Returns a new instance of this TrackingMetadata with updated info
        """
        updated = []
        for task, order in zip(self.tasks, self.orders):
            order_ = order.copy()
            task_info = get_task_metadata(task)
            order.update(task_info)
            updated.append(order)

        data = dict(self)
        data['orders'] = updated

        return self.__class__(data)

    # schema
    def validate(self):
        validation.validate_with_defaults(self._metadata, self.schema,
                                          resolve=SCHEMA_TRACKING.parent)

    @staticmethod
    def _load_schema(filename=SCHEMA_TRACKING):
        with open(str(filename)) as src:
            return json.load(src)

    # repr, printing, etc info
    def __repr__(self):
        return repr_tracking(self)

    def _repr_html_(self):
        # TODO
        return "CEDAR Tracking Metadata"


# -----------------------------------------------------------------------------
# String repr
def repr_tracking(tracking_data,
                  show_program=True, show_submission=True, show_tracking=True,
                  show_states=True, show_runtimes=True,
                  show_orders=None):
    """ Return string formatted information about CEDAR tracking metadata
    """
    lines = []

    if show_program:
        lines.extend(repr_metadata_program(tracking_data['program']))
    if show_submission:
        lines.extend(repr_tracking_submission(tracking_data['submission']))
    if show_tracking:
        lines.extend(repr_tracking_tracking(tracking_data['tracking']))

    # Top level order information is always added
    lines.extend(repr_tracking_orders(tracking_data['orders'],
                                      show_states=show_states,
                                      show_runtimes=show_runtimes))

    if show_orders is not None:
        # Handle 'True', which is shortcut to showing all order details
        if show_orders is True:
            show_orders = range(len(tracking_data['orders']))

        # Otherwise expect sequence of order indexes (1st, 2nd, etc)
        for order_id in show_orders:
            order_md = tracking_data['orders'][order_id]
            order_repr = repr_metadata_order(order_md, f'Order #{order_id}')
            lines.extend(order_repr)

    return '\n'.join(lines)


def repr_metadata_program(info, header='Program Info:'):
    lines = _indent([
        f'* {info["name"]}={info["version"]}',
        f'* earthengine-api={info["ee"]}'
    ], n=1)
    return [header] + lines


def repr_tracking_submission(info, header='Submission Info:'):
    tile_indices = [f'({i[0]}, {i[1]})' for i in info['tile_indices']]

    # Format subsection first so it can be added as a string
    period_info_str = ['* Period:'] + _indent([
        f'* Start: {info["period_start"]}',
        f'* End:   {info["period_end"]}',
        f'* Freq:  {info["period_freq"]}'
    ], n=2)

    lines = _indent([
        f'* Submitted on {info["submitted"]}',
        f'* Tile Grid: "{info["tile_grid"]["name"]}"',
        f'* Tile Indices : {", ".join(tile_indices)}',
        '\n'.join(period_info_str)
    ])
    return [header] + lines


def repr_tracking_tracking(info, header='Tracking Info'):
    lines = _indent([
        f'* Name: {info["name"]}',
        f'* Prefix: {info["prefix"]}',
        f'* Collections: {", ".join(info["collections"])}',
        f'* Image template: {info["name_template"]}',
        f'* Image prefix: {info["prefix_template"]}'
    ], n=1)
    return [header] + lines


def repr_tracking_orders(info, show_states=True, show_runtimes=True,
                         header='Orders'):
    lines = [f'* Count: {len(info)}']
    if show_states:
        states = summarize_states(info)
        lines_ = [f'- {state}: {n}' for state, n in states.items()]
        lines.append('* States:')
        lines.extend(_indent(lines_, n=1))

    if show_runtimes:
        runtimes = summarize_runtimes(info)
        mean = _format_runtime(runtimes['mean'])
        std = _format_runtime(runtimes['std'])
        line = f'* Runtime: {mean} +/- {std} minutes'
        lines.append(line)

    return [header] + _indent(lines)


def repr_metadata_order(info, header='Order:'):
    summary = _summarize_order(info)
    s_runtime = _format_runtime(summary['runtime'])
    lines = _indent([
        f'- Name: {summary["name"]}',
        f'- Prefix: {summary["prefix"]}',
        f'- ID: {summary["id"]}',
        f'- Task state: {summary["state"]}',
        f'- Runtime: {s_runtime} minutes',
        f'- Image pieces: {summary["n_images"]}',
        f'- Output URL: {summary["output_url"]}',
    ], n=1)
    return [header] + lines


# -----------------------------------------------------------------------------
# Calculations
def calculate_order_runtime(start_time, update_time, nan=np.nan):
    if start_time and update_time:
        return update_time - start_time
    elif start_time:
        now = dt.datetime.now().timestamp()
        return now - start_time
    else:
        return nan


def summarize_states(orders):
    """ Returns tallies of order task status
    """
    order_states = defaultdict(lambda: 0)
    for order in orders:
        order_states[order['status']['state']] += 1
    return dict(order_states)


def summarize_runtimes(orders):
    """ Summarize runtimes of order tasks
    """
    runtimes = []
    for order in orders:
        runtime = calculate_order_runtime(
            order['status']['start_timestamp_ms'],
            order['status']['update_timestamp_ms'])
        runtimes.append(runtime)

    runtimes = np.asarray(runtimes)
    mask = np.isfinite(runtimes)
    any_ = mask.any()

    return {
        'runtimes': runtimes,
        'total': np.sum(runtimes[mask]) if any_ else np.nan,
        'n': runtimes.size,
        'mean': np.mean(runtimes[mask]) if any_ else np.nan,
        'std': np.std(runtimes[mask]) if any_ else np.nan
    }


def _summarize_order(order):
    summary = {k: order[k] for k in ('name', 'prefix', )}
    summary['id'] = order['status']['id']
    summary['state'] = order['status']['state']
    summary['runtime'] = calculate_order_runtime(
        order['status']['start_timestamp_ms'],
        order['status']['update_timestamp_ms'])
    summary['n_images'] = len(order['status']['output_url'])
    summary['output_url'] = list(set(order['status']['output_url']))

    return summary


# Helpers
def _indent(lines, n=1, length=4):
    return [' ' * n * length + line for line in lines]


def _format_runtime(time_ms):
    if time_ms:
        return '{0:02.2f}'.format(time_ms / 60. / 1000.)
    else:
        return 'nan'


_EE_UNSUBMITTED = 'UNSUBMITTED'
_EE_READY = 'READY'
_EE_RUNNING = 'RUNNING'
_EE_COMPLETED = 'COMPLETED'
_EE_FAILED = 'FAILED'
_EE_CANCEL_REQUESTED = 'CANCEL_REQUESTED'
_EE_CANCELLED = 'CANCELLED'



""" Sensor specific ARD submission tools
"""
from . import landsat


#: dict: Mapping of GEE collection to ARD creating function
CREATE_ARD_COLLECTION = {}
CREATE_ARD_COLLECTION.update({
    k: landsat.create_ard for k in landsat.METADATA.keys()
})



""" CLI for checking status of tasks and exports
"""
from collections import defaultdict
import json
import logging
from pathlib import Path

import click

from . import options


@click.group('status', short_help='Check status of tasks and exports')
@click.pass_context
def group_status(ctx):
    """ Programs for checking, printing, modifying, and cancelling orders
    """
    pass


@group_status.command('list', short_help='List tracked metadata')
@click.pass_context
def list(ctx):
    """ List tracked orders
    """
    logger = ctx.obj['logger']

    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    # Find tracker metadata
    try:
        tracked_infos = tracker.list()
    except FileNotFoundError as e:
        click.echo('No tracked orders (could not find tracking prefix)')
    else:
        # Display
        if tracked_infos:
            if logger.level <= logging.WARNING:
                click.echo('Tracked orders:')
            for tracked_info in tracked_infos:
                click.echo(tracked_info)
        else:
            click.echo('No tracked orders')


@group_status.command('cancel',
                      short_help='Cancel Earth Engine tasks for this order')
@options.arg_tracking_name
@click.pass_context
def cancel(ctx, tracking_name):
    from ..utils import get_ee_tasks, load_ee
    ee_api = load_ee(True)

    logger = ctx.obj['logger']
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    # Read tracking info
    info = tracker.read(tracking_name)

    # Cancel tasks
    for task in info.tasks:
        task.cancel()
        click.echo(f'Cancelled task ID "{task.id}"')

    # Delete tracking
    tracker.clean(info, tracking_name=tracking_name)

    if logger.level <= logging.WARNING:
        click.echo('Complete')


@group_status.command('print', short_help='Print job tracking info')
@options.arg_tracking_name
@click.option('--order', 'order_id', type=int, multiple=True,
              help='Display verbose info about a specific order')
@click.option('--all', 'all_orders', is_flag=True,
              help='Display verbose info about all orders')
@click.pass_context
def print(ctx, tracking_name, order_id, all_orders):
    """ Print job submission tracking info
    """
    from cedar.metadata import repr_tracking

    logger = ctx.obj['logger']
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    info = tracker.read(tracking_name)

    if all_orders:
        show_orders = True
    elif order_id:
        show_orders = order_id
    else:
        show_orders = None

    click.echo(repr_tracking(info, show_orders=show_orders))


@group_status.command('completed', short_help='Check if order has completed')
@click.argument('tracking_name', type=str)
@options.opt_update_order
@click.pass_context
def completed(ctx, tracking_name, update_):
    """ Print percent of order completed & exit 1 if not complete
    """
    from cedar.utils import load_ee
    ee = load_ee(True)

    logger = ctx.obj['logger']
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    if update_:
        info = tracker.update(tracking_name)
    else:
        info = tracker.read(tracking_name)

    percent = info.progress

    if logger.level <= logging.WARNING:
        click.echo(f'Percent complete: {percent * 100. :03.2f}')

    if percent == 1.0:
        ctx.exit(0)
    else:
        ctx.exit(1)


@group_status.command('update', short_help='Update tracking info')
@click.argument('tracking_name', required=False, type=str)
@click.option('--all', 'all_', is_flag=True, help='Update all tracked orders')
@click.option('--dest', type=click.Path(file_okay=False, resolve_path=True),
              help='Save a local copy of tracking information to this folder')
@click.pass_context
def update(ctx, tracking_name, all_, dest):
    """ Update job submission tracking info
    """
    if all_ and tracking_name:
        raise click.BadParameter('Cannot specify tracking names AND `--all`',
                                 param_hint='--all')

    from cedar.utils import load_ee
    ee = load_ee(True)

    logger = ctx.obj['logger']
    config = options.fetch_config(ctx)
    tracker = config.get_tracker()

    if all_:
        tracking_name = tracker.list()
    else:
        tracking_name = [tracking_name]

    dest = Path(dest) if dest else None

    for name in tracking_name:
        if logger.level <= logging.WARNING:
            click.echo(f'Updating "{name}"')

        info = tracker.update(name)

        if dest:
            dest.mkdir(exist_ok=True, parents=True)
            dest_ = dest.joinpath(name)
            with open(str(dest_), 'w') as dst:
                json.dump(info, dst, indent=2)

    if logger.level <= logging.WARNING:
        click.echo('Complete')



""" Convert "pre-ARD" to ARD
"""
from collections import defaultdict
import datetime as dt
import json
import logging
import os
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from stems.gis import convert, georeference, grids
from stems.io.encoding import netcdf_encoding

from . import defaults, __version__

logger = logging.getLogger(__name__)


def process_preard(metadata, images, chunks=None):
    """ Open and process pre-ARD data to ARD

    Parameters
    ----------
    metadata : dict
        Image metadata
    images : Sequence[str or Path]
        Path(s) to pre-ARD imagery
    chunks : dict, optional
        Chunks to use when opening pre-ARD GeoTIFF files. If ``None``,
        defaults to ``{'x': 256, 'y': 256, 'band': -1}``

    Returns
    -------
    xr.Dataset
        pre-ARD processed to (in memory) ARD format that can be written to disk
    """

    image_metadata = metadata['image']
    # Read metadata and determine key attributes
    times = pd.to_datetime([_ard_image_timestamp(images)
                            for images in image_metadata['images']]).values
    bands = image_metadata['bands']

    # Create pre-ARD DataArray
    preard_da = read_preard(images, chunks=chunks)

    # Remove any attributes
    preard_da.attrs = {}

    # Convert to Dataset
    ard_ds = preard_to_ard(preard_da, times, bands)

    # Attach attribute metadata
    version = metadata['program']['version']
    order_metadata = metadata['order']
    collection = order_metadata['collection']
    submitted = order_metadata.get('submitted', 'UNKNOWN TIME')
    date_start = order_metadata['date_start']
    date_end = order_metadata['date_end']
    dt_now = dt.datetime.today().strftime("%Y%m%dT%H%M%S")

    attrs = {
        'title': f'Collection "{collection}" Analysis Ready Data',
        'history': '\n'.join([
            (f'{submitted} - Ordered pre-ARD from GEE for collection '
             f'"{collection}" between {date_start}-{date_end} using '
             f'`cedar={version}`'),
             f'{dt_now} - Converted to ARD using `cedar={__version__}`'
        ]),
        'source': f'Google Earth Engine Collection "{collection}"',
        'images': json.dumps(image_metadata['images'])
    }
    ard_ds.attrs = attrs

    # Georeference
    tile_ = grids.Tile.from_dict(metadata['tile'])
    ard_ds = georeference(ard_ds, tile_.crs, tile_.transform)

    return ard_ds


def find_preard(path, metadata_pattern='*.json'):
    """ Match pre-ARD metadata with imagery in some location

    Parameters
    ----------
    path : str or Path
        Path to a metadata file or directory of files (returning matches
        inside the directory)

    Returns
    -------
    dict[str, list[str]]
        Pairs of metadata filename to image filename(s)
    """
    path = Path(path)
    if path.is_dir():
        metadata = list(path.glob(metadata_pattern))
    else:
        metadata = [path]

    preard = {}
    for meta in metadata:
        images = sorted(meta.parent.glob(meta.stem + '*.tif'))
        if images:
            preard[meta] = images
        else:
            logger.debug(f'Could not find images for metadata file {meta}')

    return preard


def preard_to_ard(xarr, time, bands):
    """ Convert a "pre-ARD" DataArray to an ARD xr.Dataset

    Parameters
    ----------
    xarr : xarray.DataArray
        DataArray containing observations from all bands and time
    time : np.ndarray
        Time information for each observation
    bands : Sequence[str]
        Band names

    Returns
    -------
    xr.Dataset
        Dataset containing all observations split into subdatasets
        according to band

    Raises
    ------
    ValueError
        Raised if the number of bands and times specified do not match
        the number of "bands" in the input DataArray
    """
    n_band = len(bands)
    n_time = len(time)
    n_band_time = n_band * n_time

    if n_band * n_time != xarr.band.size:
        raise ValueError('Number of bands x time specified does not match '
                         'input data ({xarr.band.size})')

    ds_bands = {}
    for i_band, band_name in enumerate(bands):
        # Select out this band from list of band x time
        da_band = xarr[np.arange(i_band, n_band_time, n_band), ...]
        # Replace "band" for "time" in two steps
        da_band.coords['time'] = ('band', time)
        ds_bands[band_name] = da_band.swap_dims({'band': 'time'}).drop('band')

    ard_ds = xr.Dataset(ds_bands)
    return ard_ds


def ard_netcdf_encoding(ard_ds, metadata, **encoding_kwds):
    """ Return encoding for ARD NetCDF4 files

    Parameters
    ----------
    ard_ds : xr.Dataset
        ARD as a XArray Dataset
    metadata : dict
        Metadata about ARD

    Returns
    -------
    dict
        NetCDF encoding to use with :py:meth:`xarray.Dataset.to_netcdf`
    """
    assert 'nodata' not in encoding_kwds
    nodata = metadata['image'].get('nodata', None)
    encoding = netcdf_encoding(ard_ds, nodata=nodata, **encoding_kwds)
    return encoding


def read_metadata(filename):
    """ Read pre-ARD image metadata from a file
    """
    with open(filename, 'r') as f:
        meta = json.load(f)
    return meta


def read_preard(filenames, chunks=None):
    """ Read pre-ARD file(s) into a single DataArray

    Parameters
    ----------
    filenames : Sequence[str or Path]
        Pre-ARD file name(s)
    chunks : dict, optional
        Chunks to use when opening pre-ARD GeoTIFF files. If ``None``,
        defaults to ``{'x': 256, 'y': 256, 'band': -1}``

    Returns
    -------
    xr.DataArray
        Pre-ARD joined together as a single DataArray
    """
    if isinstance(filenames, (str, Path)):
        filenames = (filenames, )
    filenames = [Path(f) for f in filenames]

    if chunks is None:
        chunks = defaults.PREARD_CHUNKS.copy()

    common = os.path.commonprefix([f.stem for f in filenames])

    by_row = defaultdict(list)
    for fname in filenames:
        shard = fname.stem[len(common):]
        if not shard:  # OK if just 1 file
            assert len(filenames) == 1
            ymin, xmin = 0, 0
        else:
            ymin, xmin = map(int, shard.split('-'))
        da = xr.open_rasterio(fname, chunks=chunks)
        by_row[ymin].append(da)

    # Concat across columns
    rows = {row: xr.concat(by_row[row], dim='x') for row in by_row}

    # Concat across rows
    preard = xr.concat(rows.values(), dim='y')

    return preard


def _ard_image_timestamp(images):
    return _unix2dt(min(i['system:time_start'] for i in images))


def _unix2dt(timestamp):
    return dt.datetime.fromtimestamp(timestamp / 1e3)



"""Main group of commands for cedar CLI
"""
import logging
from pkg_resources import iter_entry_points

import click
import click_plugins

import cedar

from . import options


_context = dict(
    token_normalize_func=lambda x: x.lower(),
    help_option_names=['--help', '-h']
)


@click_plugins.with_plugins(ep for ep in
                            iter_entry_points('cedar.cli'))
@click.group(help='cedar command line interface', context_settings=_context)
@options.opt_config_file
@click.version_option(cedar.__version__)
@click.option('--verbose', '-v', count=True, help='Be verbose')
@click.option('--quiet', '-q', count=True, help='Be quiet')
@click.pass_context
def main(ctx, config_file, verbose, quiet):
    """ cedar

    Creator for Analysis Ready Data (ARD)

    Home: https://github.com/ceholden/cedar
    """
    # Setup logging
    from stems.logging_ import setup_logger
    verbosity = verbose - quiet
    log_level = max(logging.DEBUG, logging.WARNING - verbosity * 10)
    logger = setup_logger('cedar', level=log_level)

    # Setup config file (if provided)
    if config_file:
        from cedar.config import Config
        config = Config.from_yaml(config_file)
    else:
        config = None

    ctx.obj = {}
    ctx.obj['logger'] = logger
    ctx.obj['config_file'] = config



""" cedar CLI options and defaults
"""
import logging
import os

import click

from .. import defaults


# =============================================================================
# Arguments
arg_tracking_name = click.argument('tracking_name', type=str)


# Types
type_path_file = click.Path(exists=True, dir_okay=False, resolve_path=True)

# Options
opt_browser = click.option(
    '--browser', is_flag=True,
    help='Open a web browser instead of terminal to authenticate')
opt_overwrite = click.option('--overwrite', is_flag=True,
                             help='Overwrite existing files')


opt_update_order = click.option(
    '--update', 'update_', is_flag=True,
    help='Update the tracking info before continuing')


# =============================================================================
# Print styling
def _click_style(bg, fg, **kwds_):
    def inner_(*args, **kwds):
        kwds.update(kwds_)
        return click.style(*args, bg=bg, fg=fg, **kwds)
    return inner_

STYLE_ERROR = _click_style(None, 'red')
STYLE_WARNING = _click_style(None, 'red')
STYLE_INFO = _click_style(None, None)
STYLE_DEBUG = _click_style(None, 'cyan')


# =============================================================================
# Configuration file
def fetch_config(ctx, fail_if_missing=True):
    """ Fetch ``config`` from a click context with error handling

    Parameters
    ----------
    ctx : click.Context
        Click CLI context
    fail_if_missing : bool, optional
        Raise an exception if config is missing. Otherwise returns
        None

    Returns
    -------
    dict
        Configuration data

    Raises
    ------
    click.BadParameter
        Raised if configuration isn't available
    """
    config_file = ctx.obj and ctx.obj.get('config_file', None)

    if config_file:
        return config_file
    elif fail_if_missing:
        ctx_, param = fetch_param(ctx, 'config_file')
        msg = ('Must specify configuration file. Pass it to the program as '
               'an option to the main program (`cedar -C CONFIG_FILE ...`) '
               'or using the environment variable '
               f'(`{defaults.ENVVAR_CONFIG_FILE}=CONFIG_FILE cedar ...`)')
        raise click.BadParameter(msg, ctx=ctx_, param=param)
    else:
        return {}


opt_config_file = click.option(
    '--config_file', '-C',
    default=lambda: os.environ.get(defaults.ENVVAR_CONFIG_FILE, None),
    allow_from_autoenv=True,
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help='Configuration file'
)


# HELPERS
def fetch_param(ctx, name):
    """ Try to fetch a click.Parameter from a click.Context (or its parents)
    """
    # Try to raise error
    parent = ctx
    while parent is not None:
        params = {p.name: p for p in parent.command.params}
        if name in params:
            return parent, params[name]
        else:
            parent = getattr(parent, 'parent', None)
    return None, None



""" Default values for cedar
"""
import os


# =============================================================================
# Configuration file defaults
# str: Environment variable used for passing configuration file location
ENVVAR_CONFIG_FILE = 'CEDAR_CONFIG_FILE'

# Sequence[str]: Paths to cedar user configuration data
CEDAR_ROOT_CONFIG = [
    os.path.join(os.path.expanduser('~'), '.config', 'cedar'),
    os.getcwd()
]
if 'CEDAR_ROOT_CONFIG' in os.environ:
    CEDAR_ROOT_CONFIG.insert(0, os.environment['CEDAR_ROOT_CONFIG'])


# =============================================================================
# Google Earth Engine
#: str: Default pre-ARD collection frequency (default: 1 year (Jan 1 - Dec 31))
PREARD_FREQ = '1YS'

#: str: Default pre-ARD image export name template
PREARD_NAME = "{collection}_h{tile.horizontal:03d}v{tile.vertical:03d}_{date_start}_{date_end}"
#: str: Default pre-ARD image export prefix/path template
PREARD_PREFIX = "CEDAR_PREARD"
#: str: Default pre-ARD task tracking name template
PREARD_TRACKING = 'TRACKING_PERIOD{date_start}-{date_end}_TASK{today}'
#: str: Default pre-ARD task tracking prefix/path
PREARD_TRACKING_PREFIX = 'CEDAR_TRACKING'


EXPORT_IMAGE_STRFTIME = '%Y%m%d'
EXPORT_TRACK_STRFTIME = '%Y%m%dT%H%M%S'


# =============================================================================
GDRIVE_USE_APPPROPERTIES = False


# =============================================================================
# Pre-ARD Ingest
#: dict: Chunks to use when opening Pre-ARD images
PREARD_CHUNKS = {'y': 256, 'x': 256, 'band': -1}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from setuptools import find_packages, setup

import versioneer


DISTNAME = 'cedar-datacube'
VERSION = versioneer.get_version()
CMDCLASS = versioneer.get_cmdclass()
LICENSE="BSD license",

AUTHOR = "Chris Holden"
AUTHOR_EMAIL = 'ceholden@gmail.com'

URL = 'https://github.com/ceholden/cedar-datacube'
DESCRIPTION = "Creator for Analysis Ready Data (ARD)"
CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: BSD License",
    'Natural Language :: English',
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Topic :: Software Development :: Libraries",
    "Topic :: Scientific/Engineering :: GIS",
]

with open('README.rst') as f:
    LONG_DESCRIPTION = f.read()
with open('CHANGELOG.rst') as f:
    CHANGELOG = f.read()


ENTRY_POINTS = {
    'console_scripts': [
        'cedar=cedar.cli.main:main'
    ],
    'cedar.cli': [
        'auth=cedar.cli.auth:group_auth',
        'clean=cedar.cli.storage:clean',
        'config=cedar.cli.config:group_config',
        'console=cedar.cli.console:console',
        'convert=cedar.cli.convert:convert',
        'download=cedar.cli.storage:download',
        'gee=cedar.cli.gee:group_gee',
        'status=cedar.cli.status:group_status',
        'submit=cedar.cli.submit:submit',
    ]
}


PYTHON_REQUIRES = '>=3.6'
INSTALL_REQUIRES = [
    'earthengine-api',
    'affine', 'numpy', 'pandas', 'rasterio', 'shapely', 'xarray',
    'jsonschema', 'pyyaml'
]
INSTALL_REQUIRES.extend(['click>=6.0', 'click-plugins', 'cligj>=0.5'])
if 'pytest' in sys.argv:  # only include if we're testing
    SETUP_REQUIRES = ['pytest-runner']
else:
    SETUP_REQUIRES = []
TESTS_REQUIRE = [
    'pytest', 'pytest-cov', 'pytest-lazy-fixture', 'coverage',
    'sphinx', 'sphinx_rtd_theme', 'sphinxcontrib-programoutput',
    'sphinxcontrib-bibtex', 'numpydoc',
]
EXTRAS_REQUIRE = {
    'core': INSTALL_REQUIRES,
    'tests': TESTS_REQUIRE,
    'gcs': ['google-cloud-storage'],
    'gdrive': ['google-api-python-client', 'google-auth-httplib2',
               'google-auth-oauthlib']
}
EXTRAS_REQUIRE['all'] = sorted(set(sum(EXTRAS_REQUIRE.values(), [])))


setup(
    name=DISTNAME,
    version=VERSION,
    license=LICENSE,
    description=DESCRIPTION,
    long_description='\n'.join([LONG_DESCRIPTION, CHANGELOG]),
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    classifiers=CLASSIFIERS,
    url=URL,
    packages=find_packages(),
    entry_points=ENTRY_POINTS,
    zip_safe=False,
    include_package_data=True,
    python_requires=PYTHON_REQUIRES,
    install_requires=INSTALL_REQUIRES,
    setup_requires=SETUP_REQUIRES,
    tests_require=TESTS_REQUIRE,
    extras_require=EXTRAS_REQUIRE
)



""" Tracker to submit and download GEE pre-ARD tasks
"""
from collections import defaultdict
import datetime as dt
import itertools
import logging
import os
from pathlib import Path
import string

import ee
import pandas as pd

from stems.gis.grids import TileGrid, Tile

from . import defaults, ordering, utils
from .exceptions import EmptyCollectionError
from .metadata import TrackingMetadata, get_submission_info

logger = logging.getLogger(__name__)

_STR_FORMATTER = string.Formatter()


class Tracker(object):
    """ CEDAR "pre-ARD" order tracker

    Parameters
    ----------
    tile_grid : stems.gis.grids.TileGrid
        Tile Grid to use for ARD
    store : cedar.stores.Store
        A Store that can be used to store images & metadata
    name_template : str, optional
        Template for "pre-ARD" image and metadata name
    prefix_template : str, optional
        Template for "pre-ARD" image and metadata prefix
    tracking_template : str, optional
        Template for order tracking file name
    tracking_prefix : str, optional
        Order tracking file prefix folder
    filters : Dict[str, Sequence[dict or ee.Filter]]
        Earth Engine filters to apply, organized by image collection name.
        Values should either be ``ee.Filter`` objects or dictionaries
        that describe the filter (see :py:func:`cedar.utils.serialize_filter`)
    """
    def __init__(self, tile_grid, store,
                 name_template=defaults.PREARD_NAME,
                 prefix_template=defaults.PREARD_PREFIX,
                 tracking_template=defaults.PREARD_TRACKING,
                 tracking_prefix=defaults.PREARD_TRACKING_PREFIX,
                 filters=None,
                 export_image_kwds=None):
        assert isinstance(tile_grid, TileGrid)
        self.tile_grid = tile_grid
        self.store = store
        self.name_template = name_template
        self.prefix_template = prefix_template
        self.tracking_template = tracking_template
        self.tracking_prefix = tracking_prefix
        self._filters = filters or defaultdict(list)
        self.export_image_kwds = export_image_kwds or {}

    @property
    def filters(self):
        """ list[ee.Filter]: Earth Engine filters to apply
        """
        # Convert from dict as needed
        from .utils import create_filters
        return {
            image_collection: create_filters(filters)
            for image_collection, filters in self._filters.items()
        }

    def submit(self, collections, tile_indices,
               period_start, period_end, period_freq=None):
        """ Submit and track GEE pre-ARD tasks

        Parameters
        ----------
        collections: str or Sequence[str]
            GEE image collection name(s)
        tile_indices : Sequence[(int, int)]
            Tuple(s) of rows/columns in TileGrid to process
        period_start : dt.datetime
            Starting period date
        period_end : dt.datetime
            Ending period date
        period_freq : str, optional
            If provided, ``period_start``, ``period_end``, and ``period_freq``
            are interpeted as the range for :py:func:`pandas.date_range` and
            one or more Tasks will be submitted

        Returns
        -------
        str
            Task tracking information name
        str
            Task tracking information identifier (an ID, path, etc)
        """
        # TODO: add callback (e.g., for progressbar)
        # TODO: eventually allow start/end to be None (use limits of data)
        if isinstance(collections, str):
            collections = (collections, )
        assert len(tile_indices) >= 1
        if isinstance(tile_indices[0], int):
            tile_indices = [tile_indices]

        # Get tiles
        tiles = [self.tile_grid[index] for index in tile_indices]

        # Split up period into 1 or more sub-periods if freq is given
        periods = _parse_date_freq(period_start, period_end, period_freq)
        logger.debug(f'Creating {len(periods)} ARD slice(s) for date range')

        # Create tracking name
        s_tile_indices = [f'h{h:03d}v{v:03d}' for v, h in tile_indices]
        namespace = {
            'collections': collections,
            'tiles': tiles,
            'tile_indices': '_'.join(s_tile_indices),
            'period_start': period_start.isoformat(),
            'period_end': period_end.isoformat(),
            'period_freq': period_freq,
            'now': dt.datetime.now().isoformat()
        }
        tracking_name = self.tracking_template.format(**namespace)

        # Create submission info
        submission_info = get_submission_info(self.tile_grid, collections,
                                              tile_indices,
                                              period_start, period_end,
                                              period_freq)

        # Determine parameters for each submission
        iter_submit = list(itertools.product(collections, tiles, periods))

        # Create an order - submits on context exit
        with ordering.Order.create_submission(
                tracking_name, self.tracking_prefix, self.store,
                name_template=self.name_template,
                prefix_template=self.prefix_template,
                submission_info=submission_info,
                export_image_kwds=self.export_image_kwds) as order:

            # Loop over product of collections, tiles, and dates
            for collection, tile, (date_start, date_end) in iter_submit:
                logger.debug(
                    f'Ordering "{collection}" - '
                    f'"h{tile.horizontal:03d}v{tile.vertical:03d} - '
                    f'{date_start.isoformat()} to {date_end.isoformat()}')

                collection_filters = self.filters.get(collection, [])
                try:
                    order.add(collection, tile, date_start, date_end,
                              filters=collection_filters)
                except EmptyCollectionError as ece:
                    logger.debug(ece)

        return order.tracking_name, order.tracking_id

    def list(self, pattern=None):
        """ Return a list of all tracking metadata

        Parameters
        ----------
        pattern : str, optional
            Search pattern for tracking info. Specify to subset to specific
            tracking info (e.g., from some date). If ``None`` provided,
            looks for tracking information matching
            :py:attr:`~Tracker.tracking_template`

        Returns
        -------
        list[str]
            Name of stored tracking information
        """
        if pattern is None:
            d = defaultdict(lambda: '*')
            pattern = self.tracking_template.format_map(d).split('*')[0]
        return self.store.list(path=self.tracking_prefix, pattern=pattern)

    def read(self, name):
        """ Returns stored tracking information as dict

        Parameters
        ----------
        name : str
            Name of tracking metadata (e.g., taken from running
            :func:`~Tracker.list_tracking`)

        Returns
        -------
        dict
            JSON tracking info data as a dict
        """
        data = self.store.read_metadata(name, path=self.tracking_prefix)
        return TrackingMetadata(data)

    def update(self, name):
        """ Refresh and reupload tracking information by checking with the GEE

        Parameters
        ----------
        name : str
            Name of tracking metadata (e.g., taken from running
            :func:`~Tracker.list_tracking`)

        Returns
        -------
        dict
            JSON tracking info data as a dict
        """
        tracking_info = self.read(name)
        updated = tracking_info.update()
        name_ = self.store.store_metadata(dict(updated), name)
        return updated

    def download(self, tracking_info, dest, overwrite=True, callback=None):
        """ Download "pre-ARD" and metadata to a directory

        Parameters
        ----------
        tracking_info : dict
            JSON tracking info data as a dict
        dest : str or pathlib.Path
            Destination download directory
        overwrite : bool, optional
            Overwrite existing downloaded data
        callback : callable
            Callback function to execute after each file is downloaded.
            Should take arguments "item" and "n_steps". Use this for
            progress bars or other download status reporting

        Returns
        -------
        tuple[str, list[str]]
            Key value pairs mapping GEE task IDs to the filenames of
            downloaded data. Wrap it in a ``dict`` to make it not lazy
        """
        logger.debug(f'Downloading for {len(tracking_info["orders"])} tasks')
        iter_download = download_tracked(tracking_info, self.store, dest,
                                         overwrite=overwrite)

        downloaded = defaultdict(list)
        for task_id, n_images, meta, images in iter_download:
            # Download, report (if callback), and store filenames
            logger.debug(f'Downloading output for task "{task_id}" '
                         f'({n_images or "unknown"} images)')

            for meta_ in meta:
                if callback:
                    # Metadata doesn't count as a "step"
                    callback(item=task_id, n_steps=0)
                downloaded[task_id].append(meta_)

            # We might not know how many images will be downloaded
            steps_image = 1 / n_images if n_images else 0
            for image in images:
                downloaded[task_id].append(image)
                if callback:
                    callback(item=image.stem, n_steps=steps_image)

            # Update for all downloaded images at once if we didn't know
            # a priori
            if steps_image == 0 and callback:
                item = os.path.commonprefix(
                    [p.name for p in downloaded[task_id]])
                callback(item=item + '...', n_steps=1)

        return downloaded

    def clean(self, tracking_info, tracking_name=None, callback=None):
        """ Clean "pre-ARD" imagery, metadata, and tracking metadata off GCS

        Parameters
        ----------
        tracking_info : dict
            JSON tracking info data as a dict
        tracking_name : str
            Name of tracking info file (will be deleted if provided)
        callback : callable
            Callback function to execute after each file is deleted.
            Should take arguments "item" and "n_steps". Use this for
            progress bars or other download status reporting

        Returns
        -------
        dict[str, list[str]]
            Mapping of GEE Task ID to filename(s) cleaned
        """
        iter_clean = clean_tracked(tracking_info, self.store)

        cleaned = defaultdict(list)
        for task_id, n_images, names in iter_clean:
            for name in names:
                if callback:
                    callback(item=task_id, n_steps=1 / n_images)
                cleaned[task_id].append(name)

        if tracking_name:
            self.store.remove(tracking_name, self.tracking_prefix)

        return cleaned


def download_tracked(tracking_info, store, dest, overwrite=False):
    """ Download stored "pre-ARD" and metadata described by tracking info

    Parameters
    ----------
    tracking_info : dict
        Tracking information
    store : cedar.stores.Store
        cedar store class
    dest : str or pathlib.Path
        Destination download directory
    overwrite : bool, optional
        Overwrite previously downoaded data, or not

    Yields
    ------
    id : str
        Task ID
    n_images : int
        Number of images to download if known, otherwise ``None``
    metadata : generator
        Generator that downloads metadata and yields filenames
    images : generator
        Generator that downloads imagery and yields filenames
    """
    dest_ = Path(str(dest))
    if not dest_.exists():
        dest_.mkdir(exist_ok=True, parents=True)
    else:
        assert dest_.is_dir()

    orders = tracking_info['orders']
    for order in orders:
        # Get info about order
        id_ = order['status']['id']
        name, prefix = order['name'], order['prefix']
        n_images = len(order.get('output_url', [])) or None

        # Retrieve image and metadata
        dst_meta = store.retrieve_metadata(dest, name, prefix,
                                           overwrite=overwrite)
        dst_imgs = store.retrieve_image(dest, name, prefix,
                                        overwrite=overwrite)

        yield (id_, n_images, dst_meta, dst_imgs, )


def clean_tracked(tracking_info, store):
    """ Delete stored "pre-ARD" and metadata described by tracking info

    Parameters
    ----------
    tracking_info : dict
        Tracking information
    store : cedar.stores.gcs.GCSStore or cedar.stores.gdrive.GDriveStore
        cedar store class

    Yields
    ------
    id : str
        Order GEE Task ID
    names : generator
        Generator that deletes files and returns their names
    """
    orders = tracking_info['orders']
    for order in orders:
        id_ = order['status']['id']
        name, prefix = order['name'], order['prefix']
        logger.debug(f'Deleting image and metadata for id={id_}, '
                     f'name="{name}", prefix="{prefix}"')

        # Retrieve image and metadata
        names = store.list(path=prefix, pattern=name)
        yield (id_, len(names), (store.remove(name) for name in names))


def _parse_date_freq(start, end, freq=None):
    import pandas as pd  # hiding because it can be expensive to import
    start_ = pd.to_datetime(start).to_pydatetime()
    end_ = pd.to_datetime(end).to_pydatetime()
    if freq is None:
        return list(zip([start], [end]))
    else:
        # Add offset to make inclusive of end date
        from pandas.tseries.frequencies import to_offset
        offset = to_offset(freq)
        times = pd.date_range(start, end + offset, freq=freq).to_pydatetime()
        return list(zip(times[:-1], times[1:]))



""" Testing utilities
"""
from distutils import version
import importlib

import pytest


# Tools for skipping tests
def loose_version(vstring):
    # get rid of any vcs modifier
    vstring_ = vstring.split('+')[0]
    return version.LooseVersion(vstring_)


def importorskip(name, minversion=None):
    try:
        mod = importlib.import_module(name)
        has = True
        if minversion is not None:
            if loose_version(mod.__version__) < loose_version(minversion):
                raise ImportError('Version of "{modname}" not satisfied')
    except ImportError:
        has = False
    func = pytest.mark.skipif(not has, reason=f'requires {name}')
    return has, func


has_earthengine, requires_earthengine = importorskip('ee')
has_gcs, requires_gcs = importorskip('google.cloud.storage')
has_gdrive, requires_gdrive = importorskip('googleapiclient.discovery')



""" Metadata
"""
import contextlib
import datetime as dt
import json
import logging
from typing import List, Set

import ee
from stems.gis.grids import Tile

from . import __version__
from . import defaults
from .exceptions import EmptyCollectionError, EmptyOrderError
from .metadata import (TrackingMetadata,
                       get_order_metadata,
                       get_program_metadata,
                       get_task_metadata,
                       get_tracking_metadata)

from .sensors import CREATE_ARD_COLLECTION

logger = logging.getLogger(__name__)


class Order(object):
    """  CEDAR order data

    Parameters
    ----------
    tracking_name : str
        Name of the order, which is also used as the name for tracking metadata
    tracking_prefix : str
        Directory or prefix location in storage for tracking data
    name_template : str
        String format template for pre-ARD image and metadata names
    prefix_template : str
        String format template for directory or prefix of pre-ARD
    """
    def __init__(self, tracking_name, tracking_prefix,
                 name_template=None, prefix_template=None):
        self.tracking_name = tracking_name
        self.tracking_prefix = tracking_prefix
        self.name_template = name_template or defaults.PREARD_NAME
        self.prefix_template = prefix_template or defaults.PREARD_PREFIX
        self._items = []

    def __len__(self):
        return len(self._items)

    @classmethod
    @contextlib.contextmanager
    def create_submission(cls, tracking_name, tracking_prefix, store,
                          name_template=None, prefix_template=None,
                          submission_info=None, export_image_kwds=None):
        """ Create and submit an order in a single context
        """
        instance = cls(tracking_name, tracking_prefix,
                       name_template=name_template,
                       prefix_template=prefix_template)
        try:
            yield instance
        except Exception as e:
            raise e
        finally:
            instance.submit(store,
                            submission_info=submission_info,
                            export_image_kwds=export_image_kwds)

    @property
    def collections(self) -> Set[str]:
        return set([item['collection'] for item in self._items])

    @property
    def tiles(self) -> Set[Tile]:
        return set([item['tile'] for item in self._items])

    def add(self, collection, tile, date_start, date_end, filters=None):
        """ Add a "pre-ARD" image to the order
        """
        try:
            # Determine which function should be used for ARD generation
            func_create_ard = CREATE_ARD_COLLECTION[collection]
            logger.debug(f'Using function {func_create_ard}')
        except KeyError as ke:
            raise KeyError('Unknown or unsupported image collection '
                           f'"{collection}".')

        # Create image & metadata
        # Image is still "unbounded", but will be given crs, transform,
        # and size on export
        # ``image_metadata`` should have "bands", "nodata", "images"
        image, image_metadata = func_create_ard(
            collection, tile, date_start, date_end, filters=filters)

        # Create name/prefix from templates
        namespace = {
            'collection': collection.replace('/', '_'),
            'tile': tile,
            'date_start': date_start.date().isoformat(),
            'date_end': date_end.date().isoformat(),
            'now': dt.datetime.now().isoformat()
        }
        name = self.name_template.format(**namespace)
        prefix = self.prefix_template.format(**namespace)

        # Add in tile and order metadata
        self._items.append({
            'collection': collection,
            'tile': tile,
            'name': name,
            'prefix': prefix,
            'image': image,
            'metadata': image_metadata,
            'date_start': date_start,
            'date_end': date_end,
            'filters': filters
        })

    def submit(self, store, submission_info=None, export_image_kwds=None):
        """ Submit "pre-ARD" for a collection and tile to be processed

        Parameters
        ----------
        store : cedar.stores.GDriveStore or cedar.stores.GCSStore
            Storage backend to use
        submission_info : dict, optional
            Information to include in tracking metadata about the submission
        export_image_kwds : dict, optional
            Additional keywords to pass onto ``store.store_image``

        Returns
        -------
        tracking_id : str
            Tracking metadata object identitifer (an object ID for Google Drive
            or a remote path for GCS)
        """
        if not self._items:
            raise EmptyOrderError(
                'No items in order to submit (see ``Order.add``)'
            )

        # Check to make sure names are unique
        self._validate_names()

        # Submit items to order
        submitted = []
        for item in self._items:
            # Create metadata about order and submit
            order_metadata = get_order_metadata(
                item['collection'],
                item['date_start'], item['date_end'],
                item['filters']
            )
            task, task_metadata, task_metadata_id = submit_preard_task(
                item['image'], item['metadata'],
                item['name'], item['prefix'],
                item['tile'],
                store,
                order_info=order_metadata,
                start=False,
                export_image_kwds=export_image_kwds
            )
            submitted.append((task, task_metadata, task_metadata_id))

        # Create submission metadata
        # TODO: Use cedar.metadata.TrackingMetadata
        data = {
            'program': get_program_metadata(),
            'submission': submission_info or {},
            'tracking': get_tracking_metadata(self.tracking_name,
                                              self.tracking_prefix,
                                              self.name_template,
                                              self.prefix_template,
                                              self.collections,
                                              self.tiles),
            'orders': [sub[1]['task'] for sub in submitted],
            'metadata': [sub[2] for sub in submitted]
        }
        self.tracking_metadata = TrackingMetadata(data)

        # Start tasks and save tracking metadata
        for task, _, _ in submitted:
            task.start()

        self.tracking_id = store.store_metadata(dict(self.tracking_metadata),
                                                self.tracking_name,
                                                self.tracking_prefix)
        return self.tracking_id

    def _validate_names(self):
        """ Raise ValueError if there is a name collision
        """
        names = [item['name'] for item in self._items]
        if len(set(names)) != len(names):
            raise ValueError(
                'Pre-ARD items in this order do not have unique names and '
                'overwrite each other. Please make sure '
                '``self.name_template`` has enough information (collection, '
                'date_start, date_end, tile, etc) to generate unique names.')


def submit_preard_task(image, image_metadata, name, prefix, tile, store,
                       order_info=None,
                       start=False, export_image_kwds=None):
    """ Submit an EE pre-ARD processing task and store task info
    """
    # Prepare 
    export_image_kwds = export_image_kwds or {}
    export_image_kwds.update(tile_export_image_kwds(tile))

    # Create EE task
    task = store.store_image(image, name, prefix, **export_image_kwds)
    if start:
        task.start()

    # TODO: create model for PreARDMetadata and use it
    # Finish creating metadata for task
    metadata = {
        'program': get_program_metadata(),
        'order': order_info or {},
        'tile': tile.to_dict(),
        'store': {
            'service': store.__class__.__name__,
            'export_image_kwds': export_image_kwds
        },
        'task': get_task_metadata(task),
        'image': image_metadata,
    }
    metadata_id = store.store_metadata(metadata, name, prefix)

    return task, metadata, metadata_id


def tile_export_image_kwds(tile):
    """ Returns keywords to tile exported ``ee.Image`` by reprojecting/clipping

    Parameters
    ----------
    tile : stems.gis.grids.Tile
        A tile

    Returns
    -------
    dict[str, str]
        Keywords "crs", "crs_transform", and "dimensions"
    """
    kwds = {
        'crs': tile.crs.wkt,
        'crs_transform': json.dumps(tile.transform[:6]),
        'dimensions': f'{tile.width}x{tile.height}',
    }
    return kwds



# -*- coding: utf-8 -*-
""" cedar
"""
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions
__author__ = """Chris Holden"""
__email__ = 'ceholden@gmail.com'


# See: http://docs.python-guide.org/en/latest/writing/logging/
import logging  # noqa
from logging import NullHandler as _NullHandler
logging.getLogger(__name__).addHandler(_NullHandler())



""" cedar exceptions
"""


class EmptyCollectionError(ValueError):
    """ Raised when trying to create pre-ARD from an empty set of images
    """
    pass


class EmptyOrderError(ValueError):
    """ Raised when trying to submit an empty pre-ARD order
    """
    pass



""" CEDAR metadata
"""
from .core import (
    get_order_metadata,
    get_program_metadata,
    get_submission_info,
    get_task_metadata,
    get_tracking_metadata
)
from .tracking import TrackingMetadata, repr_tracking
